# encoding: utf-8

from collections import defaultdict

import six

from .internal import MISSING_OBJECT, ModelAttributeGetter, QueryPath, get_queryable_property_descriptor

__all__ = ('MISSING_OBJECT', 'get_queryable_property', 'prefetch_queryable_properties', 'reset_queryable_property')


def get_queryable_property(model, name):
    """
    Retrieve the :class:`queryable_properties.properties.QueryableProperty`
    object with the given attribute name from the given model class or raise
    an error if no queryable property with that name exists on the model class.

    :param type model: The model class to retrieve the property object from.
    :param str name: The name of the property to retrieve.
    :return: The queryable property.
    :rtype: queryable_properties.properties.QueryableProperty
    """
    return get_queryable_property_descriptor(model, name).prop


get_queryable_property.__safe_for_unpickling__ = True


def reset_queryable_property(obj, name):
    """
    Reset the cached value of the queryable property with the given name on the
    given model instance. Read-accessing the property on this model instance at
    a later point will therefore execute the property's getter again.

    :param django.db.models.Model obj: The model instance to reset the cached
                                       value on.
    :param str name: The name of the queryable property.
    """
    descriptor = get_queryable_property_descriptor(obj.__class__, name)
    descriptor.clear_cached_value(obj)


def prefetch_queryable_properties(model_instances, *property_paths):
    """
    Populate the queryable property caches for a list of model instances based
    on the given property paths.

    :param model_instances: The model instances to prefetch the property values
                            for. The instances may be objects of different
                            models as long as the given property paths are
                            valid for all of them.
    :type model_instances: collections.Sequence
    :param str property_paths: The paths to the properties whose values should
                               be fetched, which are need to be annotatable.
                               The paths may contain the lookup separator to
                               fetch values of properties on related objects
                               (make sure that the related objects are already
                               prefetched to avoid additional queries).
    """
    from ..managers import QueryablePropertiesQuerySetMixin

    # Since the model instances may be of different types and the property
    # paths may refer to properties on related objects, the first step is
    # figuring out which exact properties need to be queried and for which
    # models a query needs to be performed.
    properties_by_model = defaultdict(lambda: defaultdict(list))
    pks_by_model = defaultdict(set)
    for path in property_paths:
        query_path = QueryPath(path)
        getter = ModelAttributeGetter(query_path[:-1])
        for instance in model_instances:
            for resolved_instance in getter.get_values(instance):
                if resolved_instance is not None:
                    properties_by_model[resolved_instance.__class__][query_path[-1]].append(resolved_instance)
                    pks_by_model[resolved_instance.__class__].add(resolved_instance.pk)

    # Perform a single query for each model, querying all properties that have
    # been requested for that model (be it directly or via relations).
    for model, property_mappings in six.iteritems(properties_by_model):
        queryset = QueryablePropertiesQuerySetMixin.inject_into_object(model._base_manager.all())
        queryset = queryset.filter(pk__in=pks_by_model[model]).select_properties(*property_mappings)
        for result in queryset.values('pk', *property_mappings):
            pk = result.pop('pk')
            for property_name, value in six.iteritems(result):
                descriptor = get_queryable_property_descriptor(model, property_name)
                for instance in property_mappings[property_name]:
                    # Only populate the cache for the concrete objects the
                    # property values were requested for (different relations
                    # may lead to the same model).
                    if instance.pk == pk:
                        descriptor.set_cached_value(instance, value)



# -*- coding: utf-8 -*-

from django.db.models import Q

from ..utils.internal import QueryPath
from .base import QueryableProperty
from .mixins import AnnotationGetterMixin, BooleanMixin


class AnnotationProperty(AnnotationGetterMixin, QueryableProperty):
    """
    A property that is based on a static annotation that is even used to
    provide getter values.
    """

    def __init__(self, annotation, **kwargs):
        """
        Initialize a new property that gets its value by retrieving an
        annotated value from the database.

        :param annotation: The static annotation to use to determine the value
                           of this property.
        """
        super(AnnotationProperty, self).__init__(**kwargs)
        self.annotation = annotation

    def get_annotation(self, cls):
        return self.annotation


class AggregateProperty(AnnotationProperty):
    """
    A property that is based on an aggregate that is used to provide both
    queryset annotations as well as getter values.
    """

    def __init__(self, aggregate, **kwargs):
        """
        Initialize a new property that gets its value by retrieving an
        aggregated value from the database.

        :param django.db.models.Aggregate aggregate: The aggregate to use to
                                                     determine the value of
                                                     this property.
        """
        super(AggregateProperty, self).__init__(aggregate, **kwargs)

    def get_value(self, obj):
        return self.get_queryset_for_object(obj).aggregate(**{self.name: self.annotation})[self.name]


class RelatedExistenceCheckProperty(BooleanMixin, AnnotationGetterMixin, QueryableProperty):
    """
    A property that checks whether related objects to the one that uses the
    property exist in the database and returns a corresponding boolean value.

    Supports queryset filtering and ``CASE``/``WHEN``-based annotating.
    """

    def __init__(self, relation_path, negated=False, **kwargs):
        """
        Initialize a new property that checks for the existence of related
        objects.

        :param str relation_path: The path to the object/field whose existence
                                  is to be checked. May contain the lookup
                                  separator (``__``) to check for more remote
                                  relations.
        """
        super(RelatedExistenceCheckProperty, self).__init__(**kwargs)
        self.query_path = QueryPath(relation_path) + 'isnull'
        self.negated = negated

    @property
    def _base_condition(self):
        """
        Return the base condition for the existence check that can be applied to a queryset of the model this property
        is defined on for the positive (existence check rather than non-existence check) case.

        :return: The base condition for this property.
        :rtype: django.db.models.Q
        """
        return self.query_path.build_filter(False)

    def get_value(self, obj):
        condition = self._base_condition
        if self.negated:
            condition.negate()
        return self.get_queryset_for_object(obj).filter(condition).exists()

    def _get_condition(self, cls):
        # Perform the filtering via a subquery to avoid any side-effects that may be introduced by JOINs.
        condition = Q(pk__in=self.get_queryset(cls).filter(self._base_condition))
        if self.negated:
            condition.negate()
        return condition



# -*- coding: utf-8 -*-

from .base import QueryableProperty
from .mixins import SubqueryMixin


class SubqueryFieldProperty(SubqueryMixin, QueryableProperty):
    """
    A property that returns a field value contained in a subquery, extracting
    it from the first row of the subquery's result set.
    """

    def __init__(self, queryset, field_name, output_field=None, **kwargs):
        """
        Initialize a new property that returns a field value from a subqery.

        :param queryset: The internal queryset to use as the subquery or a
                         callable without arguments that generates the internal
                         queryset.
        :type queryset: django.db.models.QuerySet | function
        :param str field_name: The name of the subquery field whose value
                               should be returned. May refer to an annotated
                               field or queryable property inside the subquery.
        :param output_field: The output field to use for the subquery
                             expression. Only required in cases where Django
                             cannot determine the field type on its own.
        :type output_field: django.db.models.Field | None
        """
        self.field_name = field_name
        self.output_field = output_field
        super(SubqueryFieldProperty, self).__init__(queryset, **kwargs)

    def _build_subquery(self, queryset):
        from django.db.models import Subquery

        return Subquery(queryset.values(self.field_name)[:1], output_field=self.output_field)


class SubqueryExistenceCheckProperty(SubqueryMixin, QueryableProperty):
    """
    A property that checks whether or not certain objects exist in the database
    using a custom subquery.
    """

    def __init__(self, queryset, negated=False, **kwargs):
        """
        Initialize a new property that checks for the existence of database
        records using a custom subquery.

        :param queryset: The internal queryset to use as the subquery or a
                         callable without arguments that generates the internal
                         queryset.
        :type queryset: django.db.models.QuerySet | function
        :param bool negated: Whether or not to negate the ``EXISTS`` subquery
                             (i.e. the property will return ``True`` if no
                             objects exist when using ``negated=True``).
        """
        self.negated = negated
        super(SubqueryExistenceCheckProperty, self).__init__(queryset, **kwargs)

    def _build_subquery(self, queryset):
        from django.db.models import Exists

        subquery = Exists(queryset)
        if self.negated:
            subquery = ~subquery
        return subquery



# encoding: utf-8

from __future__ import unicode_literals

from copy import copy

import six
from django.db.models import Manager
from django.db.models.query import QuerySet
from django.utils.functional import cached_property

from .compat import (
    ANNOTATION_SELECT_CACHE_NAME, ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP, MANAGER_QUERYSET_METHOD_NAME, DateQuerySet,
    DateTimeQuerySet, ModelIterable, RawModelIterable, RawQuery, ValuesListQuerySet, ValuesQuerySet, chain_query,
    chain_queryset,
)
from .exceptions import QueryablePropertyDoesNotExist, QueryablePropertyError
from .query import QUERYING_PROPERTIES_MARKER, QueryablePropertiesQueryMixin
from .utils import get_queryable_property
from .utils.internal import InjectableMixin, QueryablePropertyReference, QueryPath


class LegacyIterable(object):
    """
    Base class for queryset iterables for old Django versions to mimic the
    iterable classes of new Django versions.
    """

    def __init__(self, queryset):
        """
        Initialize a new legacy iterable for the given queryset.

        :param QuerySet queryset: The queryset to perform the database query
                                  for.
        """
        self.queryset = queryset

    def __iter__(self):
        original = super(QueryablePropertiesBaseQuerySetMixin, self.queryset)
        return getattr(original, 'iterator', original.__iter__)()


class QueryablePropertiesIterableMixin(object):
    """
    Base class for iterable mixins that handle queryable properties logic.

    Can be applied to both Django's iterable classes as well as the legacy
    iterable classes.
    """

    def __init__(self, queryset, *args, **kwargs):
        super(QueryablePropertiesIterableMixin, self).__init__(chain_queryset(queryset), *args, **kwargs)

    def __iter__(self):
        self._setup_queryable_properties()
        for obj in super(QueryablePropertiesIterableMixin, self).__iter__():
            yield self._postprocess_queryable_properties(obj)

    def _setup_queryable_properties(self):  # pragma: no cover
        """
        Set up potentially contained queryable properties to be handled
        correctly when executing the query.
        """
        pass

    def _postprocess_queryable_properties(self, obj):
        """
        Perform queryable property operations for the given loaded result
        object.

        :param obj: The object returned from the database.
        :return: The final object to return as part of the iterable.
        """
        return obj


class LegacyOrderingMixin(QueryablePropertiesIterableMixin):
    """
    A mixin for the legacy iterables that properly sets up queryable properties
    that are used for ordering.

    Old Django versions do not support ordering by annotations without
    explicitly selecting them, which is why non-selected properties used for
    ordering must be changed to selected properties.
    """

    @cached_property
    def _order_by_occurrences(self):
        """
        Cache and return a dictionary mapping queryable properties contained in
        the associated query to their indexes in the order by clause.

        Only contains property references for properties that are actually used
        for ordering.

        :return: A dictionary containing queryable property references as keys
                 and lists of indexes as values.
        :rtype: dict[queryable_properties.utils.internal.QueryablePropertyReference, list[int]]
        """
        query = self.queryset.query
        occurrences = {}
        for ref in query._queryable_property_annotations:
            annotation_name = six.text_type(ref.full_path)
            indexes = [index for index, field_name in enumerate(query.order_by)
                       if field_name in (annotation_name, '-{}'.format(annotation_name))]
            if indexes:
                occurrences[ref] = indexes
        return occurrences

    @cached_property
    def _order_by_select(self):
        """
        Cache and return a set containing the references for queryable
        properties that have to be forcibly selected as they are referenced in
        ordering while not being explicitly selected.

        :return: The set of references of properties that must be selected.
        :rtype: set[queryable_properties.utils.internal.QueryablePropertyReference]
        """
        query = self.queryset.query
        select = set()
        for ref, occurrences in six.iteritems(self._order_by_occurrences):
            annotation_name = six.text_type(ref.full_path)
            if annotation_name not in query.annotation_select and annotation_name in query.annotations:
                select.add(ref)
        return select

    def _setup_queryable_properties(self):
        super(LegacyOrderingMixin, self)._setup_queryable_properties()
        query = self.queryset.query
        select = dict(query.annotation_select)

        for property_ref in self._order_by_select:
            annotation_name = six.text_type(property_ref.full_path)
            select[annotation_name] = query.annotations[annotation_name]
        setattr(query, ANNOTATION_SELECT_CACHE_NAME, select)


class QueryablePropertiesModelIterableMixin(InjectableMixin, QueryablePropertiesIterableMixin):
    """
    A mixin for iterables that yield model instances.

    Removes the ``QUERYING_PROPERTIES_MARKER``from created model instances to
    ensure that the setters of queryable properties can be used properly.
    """

    def _setup_queryable_properties(self):
        super(QueryablePropertiesModelIterableMixin, self)._setup_queryable_properties()
        self.queryset.query._use_querying_properties_marker = True

    def _postprocess_queryable_properties(self, obj):
        obj = super(QueryablePropertiesModelIterableMixin, self)._postprocess_queryable_properties(obj)
        delattr(obj, QUERYING_PROPERTIES_MARKER)
        return obj


class LegacyModelIterable(QueryablePropertiesModelIterableMixin, LegacyIterable):
    """
    Legacy iterable class for querysets that yield model instances in Django
    versions that don't require additional ordering setup.
    """


class LegacyOrderingModelIterable(QueryablePropertiesModelIterableMixin, LegacyOrderingMixin, LegacyIterable):
    """
    Legacy iterable class for querysets that yield model instances in Django
    versions that require additional ordering setup.
    """

    def _postprocess_queryable_properties(self, obj):
        for ref in self._order_by_select:
            ref.descriptor.clear_cached_value(obj)
        return super(LegacyOrderingModelIterable, self)._postprocess_queryable_properties(obj)


class LegacyValuesIterable(LegacyOrderingMixin, LegacyIterable):
    """
    Legacy iterable class for querysets that yield value dictionaries.
    """

    def _postprocess_queryable_properties(self, obj):
        obj = super(LegacyValuesIterable, self)._postprocess_queryable_properties(obj)
        for ref in self._order_by_select:
            obj.pop(six.text_type(ref.full_path), None)
        return obj


class LegacyValuesListIterable(LegacyOrderingMixin, LegacyIterable):  # pragma: no cover
    """
    Legacy iterable class for querysets that yield value tuples.
    """

    def __init__(self, queryset, *args, **kwargs):
        super(LegacyValuesListIterable, self).__init__(queryset, *args, **kwargs)
        self.flat = queryset.flat
        self.queryset.flat = False

    @cached_property
    def _discarded_indexes(self):
        """
        Cache and return the field indexes of queryable properties that were
        only selected for ordering and must thus be discarded.

        All contained indexes will be negative, i.e. are to be interpreted as
        indexes from end of the returned rows.

        :return: A set containing the field indexes to discard.
        :rtype: set[int]
        """
        aggregate_names = list(self.queryset.query.aggregate_select)
        if self.queryset._fields:
            aggregate_names = [name for name in aggregate_names if name not in self.queryset._fields]
        aggregate_names.reverse()
        forced_names = set(six.text_type(ref.full_path) for ref in self._order_by_select)
        return {-i for i, name in enumerate(aggregate_names, start=1) if name in forced_names}

    def _postprocess_queryable_properties(self, obj):
        obj = super(LegacyValuesListIterable, self)._postprocess_queryable_properties(obj)
        obj = tuple(value for i, value in enumerate(obj, start=-len(obj)) if i not in self._discarded_indexes)
        if self.flat and len(self.queryset._fields) == 1:
            return obj[0]
        return obj


class QueryablePropertiesBaseQuerySetMixin(InjectableMixin):
    """
    Base mixin for queryable properties queryset mixins that takes care of
    injecting the :class:`QueryablePropertiesQueryMixin` into the associated
    query.
    """

    def init_injected_attrs(self):
        # To work correctly, a query using the QueryablePropertiesQueryMixin is
        # required. If the current query is not using the mixin already, it
        # will be dynamically injected into the query. That way, other Django
        # extensions using custom query objects are also supported.
        # Recent Django versions (>=3.1) have a property guarding the query
        # attribute.
        query_attr_name = '_query' if hasattr(self, '_query') else 'query'
        query = getattr(self, query_attr_name)
        chain_kwargs = {}
        if RawQuery and isinstance(query, RawQuery):
            chain_kwargs['using'] = self.db
        query = chain_query(query, **chain_kwargs)
        class_name = 'QueryableProperties' + query.__class__.__name__
        setattr(self, query_attr_name, QueryablePropertiesQueryMixin.inject_into_object(query, class_name))


class QueryablePropertiesRawQuerySetMixin(QueryablePropertiesBaseQuerySetMixin):
    """
    A mixin for Django's :class:`django.db.models.RawQuerySet` objects that
    allows to populate queryable properties in raw queries.
    """

    def __iter__(self):
        original = super(QueryablePropertiesRawQuerySetMixin, self)
        # Only recent Django versions (>= 2.1) have the iterator method.
        iterator = original.__iter__ if hasattr(original, 'iterator') else self.iterator
        for obj in iterator():
            yield obj

    def iterator(self):
        iterable_class = RawModelIterable or LegacyIterable
        for obj in QueryablePropertiesModelIterableMixin.mix_with_class(iterable_class)(self):
            yield obj


class QueryablePropertiesQuerySetMixin(QueryablePropertiesBaseQuerySetMixin):
    """
    A mixin for Django's :class:`django.db.models.QuerySet` objects that allows
    to use queryable properties in filters, annotations and update queries.
    """

    @property
    def _iterable_class(self):
        # Override the regular _iterable_class attribute of recent Django
        # versions with a property that also stores the value in the instance
        # dict, but automatically mixes the
        # QueryablePropertiesModelIterableMixin into the base class on getter
        # access if the base class yields model instances. That way, the
        # queryable properties extensions stays compatible to custom iterable
        # classes while querysets can still be pickled due to the base class
        # being in the instance dict.
        cls = self.__dict__['_iterable_class']
        if issubclass(cls, ModelIterable):
            cls = QueryablePropertiesModelIterableMixin.mix_with_class(cls, 'QueryableProperties' + cls.__name__)
        return cls

    @_iterable_class.setter
    def _iterable_class(self, value):
        self.__dict__['_iterable_class'] = value

    def _clone(self, klass=None, *args, **kwargs):
        has_iterable_class = '_iterable_class' in self.__dict__
        if not has_iterable_class:  # pragma: no cover
            # In older Django versions, the class of the queryset may be
            # replaced with a dynamically created class based on the current
            # class and the value of klass while cloning (e.g when using
            # .values()). Therefore this needs to be re-injected to be on top
            # of the MRO again to enable queryable properties functionality.
            if klass:
                klass = QueryablePropertiesQuerySetMixin.mix_with_class(klass, 'QueryableProperties' + klass.__name__)
            args = (klass,) + args
        clone = super(QueryablePropertiesQuerySetMixin, self)._clone(*args, **kwargs)
        # Since the _iterable_class property may return a dynamically created
        # class, the value of a clone must be reset to the base class.
        if has_iterable_class:
            clone._iterable_class = self.__dict__['_iterable_class']
        return clone

    def _resolve_update_kwargs(self, **kwargs):
        """
        Look for the names of queryable properties in the given keyword
        arguments for an update query and correctly resolve them into their
        actual keyword arguments.

        :param kwargs: Keyword arguments of an update query.
        :return: A dictionary containing the resolved arguments.
        :rtype: dict
        """
        original_names = set(kwargs)
        for original_name in original_names:
            try:
                prop = get_queryable_property(self.model, original_name)
            except QueryablePropertyDoesNotExist:
                continue
            if not prop.get_update_kwargs:
                raise QueryablePropertyError('Queryable property "{}" does not implement queryset updating.'
                                             .format(prop))

            # Call the method recursively since queryable properties can build
            # upon each other.
            additional_kwargs = self._resolve_update_kwargs(
                **prop.get_update_kwargs(self.model, kwargs.pop(original_name)))
            # Make sure that there are no conflicting values after resolving
            # the update keyword arguments of the queryable properties.
            for additional_name, value in six.iteritems(additional_kwargs):
                if additional_name in kwargs and kwargs[additional_name] != value:
                    raise QueryablePropertyError(
                        'Updating queryable property "{prop}" would change field "{field}", but a conflicting value '
                        'was set for this field by another queryable property or explicitly in the update arguments.'
                        .format(prop=prop, field=additional_name)
                    )
                kwargs[additional_name] = value

        return kwargs

    def select_properties(self, *names):
        """
        Add the annotations of the queryable properties with the specified
        names to this query. The annotation values will be cached in the
        properties of resulting model instances, regardless of the regular
        caching behavior of the queried properties.

        :param names: Names of queryable properties.
        :return: A copy of this queryset with the added annotations.
        :rtype: QuerySet
        """
        queryset = chain_queryset(self)
        for name in names:
            property_ref = QueryablePropertyReference(get_queryable_property(self.model, name), self.model, QueryPath())
            # A full GROUP BY is required if the query is not limited to
            # certain fields. Since only certain types of queries had the
            # _fields attribute in old Django versions, fall back to checking
            # for existing selection, on which the GROUP BY would be based.
            full_group_by = not getattr(self, '_fields', self.query.select)
            with queryset.query._add_queryable_property_annotation(property_ref, full_group_by, select=True):
                pass
        return queryset

    def iterator(self, *args, **kwargs):
        # Recent Django versions use their own iterable classes, where the
        # QueryablePropertiesModelIterableMixin will be already mixed in. In
        # older Django versions, the standalone legacy iterables are used
        # instead to perform the queryable properties processing. Exceptions
        # are legacy Date(Time)QuerySets, which don't support annotations
        # and override the ordering anyway as well as querysets that don't
        # yield model instances in Django 1.8, which doesn't require the
        # legacy ordering setup.
        if ('_iterable_class' not in self.__dict__ and
                not (isinstance(self, ValuesQuerySet) and not ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP) and
                not (DateQuerySet and isinstance(self, DateQuerySet)) and
                not (DateTimeQuerySet and isinstance(self, DateTimeQuerySet))):  # pragma: no cover
            iterable_class = LegacyOrderingModelIterable
            if not ANNOTATION_TO_AGGREGATE_ATTRIBUTES_MAP:
                iterable_class = LegacyModelIterable
            elif isinstance(self, ValuesListQuerySet):
                iterable_class = LegacyValuesListIterable
            elif isinstance(self, ValuesQuerySet):
                iterable_class = LegacyValuesIterable
            return iter(iterable_class(self))
        return super(QueryablePropertiesQuerySetMixin, self).iterator(*args, **kwargs)

    def raw(self, *args, **kwargs):
        queryset = super(QueryablePropertiesQuerySetMixin, self).raw(*args, **kwargs)
        return QueryablePropertiesRawQuerySetMixin.inject_into_object(queryset)

    def update(self, **kwargs):
        # Resolve any queryable properties into their actual update kwargs
        # before calling the base update method.
        kwargs = self._resolve_update_kwargs(**kwargs)
        return super(QueryablePropertiesQuerySetMixin, self).update(**kwargs)

    @classmethod
    def apply_to(cls, queryset):
        """
        Copy the given queryset and apply this mixin (and thus queryable
        properties functionality) to it, returning a new queryset that allows
        to use queryable property interaction.

        :param QuerySet queryset: The queryset to apply this mixin to.
        :return: A copy of the given queryset with queryable properties
                 functionality.
        :rtype: QueryablePropertiesQuerySet
        """
        return cls.inject_into_object(chain_queryset(queryset))


class QueryablePropertiesQuerySet(QueryablePropertiesQuerySetMixin, QuerySet):
    """
    A special queryset class that allows to use queryable properties in its
    filter conditions, annotations and update queries.
    """

    @classmethod
    def get_for_model(cls, model):
        """
        Get a new queryset with queryable properties functionality for the
        given model. The queryset is built using the model's default manager.

        :param model: The model class for which the queryset should be built.
        :return: A new queryset with queryable properties functionality.
        :rtype: QueryablePropertiesQuerySet
        """
        return QueryablePropertiesQuerySetMixin.inject_into_object(model._default_manager.all())


class QueryablePropertiesManagerMixin(InjectableMixin):
    """
    A mixin for Django's :class:`django.db.models.Manager` objects that allows
    to use queryable properties methods and returns
    :class:`QueryablePropertiesQuerySet` instances.
    """

    def get_queryset(self):
        queryset = getattr(super(QueryablePropertiesManagerMixin, self), MANAGER_QUERYSET_METHOD_NAME)()
        return QueryablePropertiesQuerySetMixin.inject_into_object(queryset)

    get_query_set = get_queryset

    def select_properties(self, *names):
        """
        Return a new queryset and add the annotations of the queryable
        properties with the specified names to this query. The annotation
        values will be cached in the properties of resulting model instances,
        regardless of the regular caching behavior of the queried properties.

        :param names: Names of queryable properties.
        :return: A copy of this queryset with the added annotations.
        :rtype: QuerySet
        """
        return self.get_queryset().select_properties(*names)

    @classmethod
    def apply_to(cls, manager):
        """
        Copy the given manager and apply this mixin (and thus queryable
        properties functionality) to it, returning a new manager that allows
        to use queryable property interaction.

        :param Manager manager: The manager to apply this mixin to.
        :return: A copy of the given manager with queryable properties
                 functionality.
        :rtype: QueryablePropertiesManager
        """
        manager = copy(manager)
        manager.name = '<{}_with_queryable_properties>'.format(getattr(manager, 'name', None) or 'manager')
        return cls.inject_into_object(manager)


class QueryablePropertiesManager(QueryablePropertiesManagerMixin, Manager):
    """
    A special manager class that allows to use queryable properties methods
    and returns :class:`QueryablePropertiesQuerySet` instances.
    """

    @classmethod
    def get_for_model(cls, model, using=None, hints=None):
        """
        Get a new manager with queryable properties functionality for the
        given model.

        :param model: The model class for which the manager should be built.
        :param str | None using: An optional name of the database connection
                                 to use.
        :param dict | None hints: Optional hints for the db connection.
        :return: A new manager with queryable properties functionality.
        :rtype: QueryablePropertiesManager
        """
        manager = cls()
        manager.model = model
        manager.name = '<manager_with_queryable_properties>'
        manager._db = using
        manager._hints = hints or {}
        return manager



# -*- coding: utf-8 -*-

from django.db import models
from django.utils.translation import gettext_lazy as _

from queryable_properties.properties import MappingProperty, ValueCheckProperty


class ReleaseTypeModel(models.Model):
    ALPHA = 'a'
    BETA = 'b'
    STABLE = 's'
    RELEASE_TYPE_CHOICES = (
        (ALPHA, _('Alpha')),
        (BETA, _('Beta')),
        (STABLE, _('Stable')),
    )

    release_type = models.CharField(max_length=1, choices=RELEASE_TYPE_CHOICES, default=STABLE)

    is_alpha = ValueCheckProperty('release_type', ALPHA)
    is_beta = ValueCheckProperty('release_type', BETA)
    is_stable = ValueCheckProperty('release_type', STABLE)
    is_unstable = ValueCheckProperty('release_type', ALPHA, BETA)
    release_type_verbose_name = MappingProperty('release_type', models.CharField(null=True), RELEASE_TYPE_CHOICES)

    class Meta:
        abstract = True



# -*- coding: utf-8 -*-

from itertools import chain

import six
from django.contrib.admin.options import InlineModelAdmin
from django.core.exceptions import ImproperlyConfigured
from django.db.models import F, expressions

from ..compat import checks
from ..utils.internal import InjectableMixin, QueryPath, resolve_queryable_property


class Error(getattr(checks, 'Error', object)):
    """
    Custom error class to normalize check/validation error handling across
    all supported Django version. Also takes care of producing correct
    queryable properties error IDs and raising exceptions for old-style
    validations.
    """

    def __init__(self, msg, obj, error_id):
        """
        Initialize a new error object with the given data.

        :param str msg: The error message.
        :param obj: The object which was checked/validated.
        :param int error_id: A unique ID for the error.
        """
        error_id = 'queryable_properties.admin.E{:03}'.format(error_id)
        if self.__class__.__bases__ != (object,):
            super(Error, self).__init__(msg, obj=obj, id=error_id)
        else:  # pragma: no cover
            self.msg = msg
            self.obj = obj
            self.id = error_id

    def raise_exception(self):
        """Raise an ImproperlyConfigured exception for this error."""
        raise ImproperlyConfigured('{}: ({}) {}'.format(six.text_type(self.obj), self.id, self.msg))


class QueryablePropertiesChecksMixin(InjectableMixin):
    """
    A mixin for Django's recent admin checks classes as well as old-style
    validation classes that allows to validate references to queryable
    properties in admin class attributes.
    """

    def check(self, admin_obj, *args, **kwargs):
        errors = super(QueryablePropertiesChecksMixin, self).check(admin_obj, *args, **kwargs)
        # The number of arguments differs between old and recent Django
        # versions.
        model = getattr(admin_obj, 'model', args[0] if args else None)
        errors.extend(self._check_list_select_properties(admin_obj, model))
        return errors

    def validate(self, cls, model):  # pragma: no cover
        fake_cls = self._validate_queryable_properties(cls, model)
        super(QueryablePropertiesChecksMixin, self).validate(fake_cls, model)

    def _validate_queryable_properties(self, cls, model):  # pragma: no cover
        """
        Main routine regarding queryable properties for old-style validations.

        Since validations raise an error as soon as they find a problem, they
        cannot be extended the same way checks can (by inspecting errors and
        suppressing/replacing them) as a suppressed exception could leave
        subsequent errors undetected.
        Instead, all queryable property validations are performed first and
        an exception is raised if there's a problem. If everything related to
        queryable properties validates successfully, a fake admin class is
        created on the fly, which is stripped of all queryable property
        references and can therefore be subjected to Django's standard
        validations.

        :param cls: The admin class to validate.
        :param model: The model the admin class is used for.
        :return: A fake class to be validated by Django's standard validation.
        :rtype: type
        """
        list_filter = []
        ordering = []
        errors = self._check_list_select_properties(cls, model)
        pk_name = model._meta.pk.name

        if not issubclass(cls, InlineModelAdmin):
            for i, item in enumerate(cls.list_filter or ()):
                if not callable(item):
                    prop, property_errors = self._check_list_filter_queryable_property(cls, model, item,
                                                                                       'list_filter[{}]'.format(i))
                    if prop:
                        errors.extend(property_errors)
                        # Replace a valid reference to a queryable property
                        # with a reference to the PK field so avoid Django
                        # validation errors while not changing indexes of
                        # other items.
                        item = (pk_name, item[1]) if isinstance(item, (tuple, list)) else pk_name
                list_filter.append(item)

        for i, field_name in enumerate(cls.ordering or ()):
            prop, property_errors = self._check_ordering_queryable_property(cls, model, field_name,
                                                                            'ordering[{}]'.format(i))
            if prop:
                errors.extend(property_errors)
                # Replace a valid reference to a queryable property with a
                # reference to the PK field so avoid Django validation errors
                # while not changing indexes of other items.
                field_name = pk_name
            ordering.append(field_name)

        if errors:
            errors[0].raise_exception()

        # Build a fake admin class without queryable property references to be
        # validated by Django.
        return type(cls.__name__, (cls,), {
            '__module__': cls.__module__,
            'list_filter': list_filter,
            'ordering': ordering,
        })

    def _check_queryable_property(self, obj, model, query_path, label, allow_relation=True, allow_lookups=True):
        """
        Perform common checks for a (potential) referenced queryable property.

        :param obj: The admin object or class.
        :param model: The model the admin class is used for.
        :param QueryPath query_path: The query path to the queryable property.
        :param str label: A label to use for error messages.
        :param bool allow_relation: Whether or not the queryable property
                                    should be considered valid if it is
                                    reached via relations.
        :param bool allow_lookups: Whether or not the reference to the
                                   queryable property may contain lookups.
        :return: A 2-tuple containing the resolved queryable property (if any)
                 as well as a list of check errors.
        :rtype: (queryable_properties.properties.QueryableProperty, list[Error])
        """
        errors = []
        property_ref, lookups = resolve_queryable_property(model, query_path)
        if not property_ref:
            message = '"{}" refers to "{}", which is not a queryable property.'.format(label, query_path)
            errors.append(Error(message, obj, error_id=1))
        else:
            if not property_ref.property.get_annotation:
                message = ('"{}" refers to queryable property "{}", which does not implement annotation creation.'
                           .format(label, query_path))
                errors.append(Error(message, obj, error_id=2))
            if len(query_path) > len(lookups) + 1 and not allow_relation:
                message = ('The queryable property in "{}" must not be a property on a related model (invalid item: '
                           '"{}").'.format(label, query_path))
                errors.append(Error(message, obj, error_id=3))
            if lookups and not allow_lookups:
                message = ('The queryable property in "{}" must not contain lookups/transforms (invalid item: "{}").'
                           .format(label, query_path))
                errors.append(Error(message, obj, error_id=4))
        return property_ref and property_ref.property, errors

    def _check_list_filter_queryable_property(self, obj, model, item, label):
        """
        Perform checks for a (potential) queryable property used as a list
        filter item.

        :param obj: The admin object or class.
        :param model: The model the admin class is used for.
        :param str | list | tuple item: The list filter item.
        :param str label: A label to use for error messages.
        :return: A 2-tuple containing the resolved queryable property (if any)
                 as well as a list of check errors.
        :rtype: (queryable_properties.properties.QueryableProperty, list[Error])
        """
        field_name = item[0] if isinstance(item, (tuple, list)) else item
        return self._check_queryable_property(obj, model, QueryPath(field_name), label, allow_lookups=False)

    def _check_ordering_queryable_property(self, obj, model, field_name, label):
        """
        Perform checks for a (potential) queryable property used as an ordering
        item.

        :param obj: The admin object or class.
        :param model: The model the admin class is used for.
        :param str | expressions.BaseExpression field_name: The ordering item.
        :param str label: A label to use for error messages.
        :return: A 2-tuple containing the resolved queryable property (if any)
                 as well as a list of check errors.
        :rtype: (queryable_properties.properties.QueryableProperty, list[Error])
        """
        if not isinstance(field_name, six.string_types) and hasattr(expressions, 'Combinable'):
            if isinstance(field_name, expressions.Combinable):
                field_name = field_name.asc()
            if isinstance(field_name, expressions.OrderBy) and isinstance(field_name.expression, F):
                field_name = field_name.expression.name
        if field_name.startswith('-') or field_name.startswith('+'):
            field_name = field_name[1:]
        return self._check_queryable_property(obj, model, QueryPath(field_name), label)

    def _check_list_filter_item(self, obj, *args):
        errors = super(QueryablePropertiesChecksMixin, self)._check_list_filter_item(obj, *args)
        if not errors or errors[0].id != 'admin.E116':
            return errors

        # The number of arguments differs between old and recent Django
        # versions.
        model = args[0] if len(args) > 2 else obj.model
        prop, property_errors = self._check_list_filter_queryable_property(obj, model, *args[-2:])
        return property_errors if prop else errors

    def _check_ordering_item(self, obj, *args):
        errors = super(QueryablePropertiesChecksMixin, self)._check_ordering_item(obj, *args)
        if not errors or errors[0].id != 'admin.E033':
            return errors

        # The number of arguments differs between old and recent Django
        # versions.
        model = args[0] if len(args) > 2 else obj.model
        prop, property_errors = self._check_ordering_queryable_property(obj, model, *args[-2:])
        return property_errors if prop else errors

    def _check_list_select_properties(self, obj, model):
        """
        Perform checks for the `list_select_properties` value as a whole.

        :param obj: The admin object or class.
        :param model: The model the admin class is used for.
        :return: A 2-tuple containing the resolved queryable property (if any)
                 as well as a list of check errors.
        :rtype: (queryable_properties.properties.QueryableProperty, list[Error])
        """
        if not isinstance(obj.list_select_properties, (list, tuple)):
            return [Error('The value of "list_select_properties" must be a list or tuple.', obj, error_id=5)]
        return list(chain.from_iterable(
            self._check_list_select_properties_item(obj, model, item, 'list_select_properties[{}]'.format(index))
            for index, item in enumerate(obj.list_select_properties)
        ))

    def _check_list_select_properties_item(self, obj, model, item, label):
        """
        Perform checks for a (potential) queryable property used as a
        `list_select_properties` item.

        :param obj: The admin object or class.
        :param model: The model the admin class is used for.
        :param str item: The list_select_properties item.
        :param str label: A label to use for error messages.
        :return: A 2-tuple containing the resolved queryable property (if any)
                 as well as a list of check errors.
        :rtype: (queryable_properties.properties.QueryableProperty, list[Error])
        """
        return self._check_queryable_property(obj, model, QueryPath(item), label,
                                              allow_relation=False, allow_lookups=False)[1]



# encoding: utf-8

from datetime import date

import pytest
from django import VERSION as DJANGO_VERSION
from mock import Mock

try:
    from django.db.models import Value
except ImportError:
    Value = Mock()
try:
    from django.db.models.functions import Concat
except ImportError:
    Concat = Mock(return_value=Mock(output_field=None))

from .app_management.models import (
    ApplicationTag, ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
    CategoryWithClassBasedProperties, CategoryWithDecoratorBasedProperties,
)
from .dummy_lib.models import ReleaseTypeModel

collect_ignore = []
if DJANGO_VERSION < (4, 1):
    collect_ignore.append('test_queries/test_async.py')


@pytest.fixture
def categories():
    return [
        CategoryWithClassBasedProperties.objects.create(name='Linux apps'),
        CategoryWithClassBasedProperties.objects.create(name='Windows apps'),
        CategoryWithDecoratorBasedProperties.objects.create(name='Linux apps'),
        CategoryWithDecoratorBasedProperties.objects.create(name='Windows apps'),
    ]


@pytest.fixture
def applications(categories):
    apps = [
        ApplicationWithClassBasedProperties.objects.create(name='My cool App'),
        ApplicationWithClassBasedProperties.objects.create(name='Another App'),
        ApplicationWithDecoratorBasedProperties.objects.create(name='My cool App'),
        ApplicationWithDecoratorBasedProperties.objects.create(name='Another App'),
    ]
    apps[0].categories.add(categories[0])
    apps[1].categories.add(categories[0])
    apps[1].categories.add(categories[1])
    apps[2].categories.add(categories[2])
    apps[3].categories.add(categories[2])
    apps[3].categories.add(categories[3])
    return apps


@pytest.fixture
def versions(applications):
    objs = []
    for application in applications:
        objs.extend([
            application.versions.create(major=1, minor=2, patch=3, release_type=ReleaseTypeModel.BETA,
                                        supported_until=date(2016, 12, 31)),
            application.versions.create(major=1, minor=3, patch=0,
                                        supported_from=date(2017, 1, 1), supported_until=date(2017, 12, 31)),
            application.versions.create(major=1, minor=3, patch=1,
                                        supported_from=date(2018, 1, 1), supported_until=date(2018, 12, 31)),
            application.versions.create(major=2, minor=0, patch=0, changes='Amazing new features',
                                        release_type=ReleaseTypeModel.ALPHA, supported_from=date(2018, 11, 1)),
        ])
    return objs


@pytest.fixture
def tags(applications):
    tags = [
        ApplicationTag.objects.create(label='FOSS'),
        ApplicationTag.objects.create(label='Proprietary'),
    ]
    applications[0].tags.add(tags[0])
    applications[1].tags.add(tags[1])
    return tags



# -*- coding: utf-8 -*-
import pytest
import six
from django import VERSION as DJANGO_VERSION
from django.db.models import Avg, Q

from queryable_properties.properties import AggregateProperty, AnnotationProperty, RelatedExistenceCheckProperty
from queryable_properties.utils import get_queryable_property
from queryable_properties.utils.internal import QueryPath
from ..app_management.models import ApplicationWithClassBasedProperties, CategoryWithClassBasedProperties

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


class TestAnnotationProperty(object):

    @pytest.mark.parametrize('cached, expected_cached', [
        (None, False),
        (False, False),
        (True, True),
    ])
    def test_initializer(self, cached, expected_cached):
        annotation = Avg('test')
        prop = AnnotationProperty(annotation, cached=cached)
        assert prop.annotation is annotation
        assert prop.cached is expected_cached

    def test_getter(self, categories):
        assert categories[0].version_count == 8
        assert categories[1].version_count == 4

    def test_annotation(self, categories):
        assert list(CategoryWithClassBasedProperties.objects.order_by('-version_count')) == [
            categories[0], categories[1]
        ]


class TestAggregateProperty(object):

    @pytest.mark.parametrize('cached, expected_cached', [
        (None, False),
        (False, False),
        (True, True),
    ])
    def test_initializer(self, cached, expected_cached):
        annotation = Avg('test')
        prop = AggregateProperty(annotation, cached=cached)
        assert prop.annotation is annotation
        assert prop.cached is expected_cached

    def test_getter(self, applications, versions):
        assert applications[0].major_sum == 5
        versions[0].delete()
        assert applications[0].major_sum == 4


class TestRelatedExistenceCheckProperty(object):

    @pytest.mark.parametrize('path, kwargs, expected_query_path', [
        ('my_field', {}, QueryPath('my_field__isnull')),
        ('my_relation__my_field', {'negated': True}, QueryPath('my_relation__my_field__isnull')),
        ('my_field', {'cached': True}, QueryPath('my_field__isnull')),
    ])
    def test_initializer(self, path, kwargs, expected_query_path):
        prop = RelatedExistenceCheckProperty(path, **kwargs)
        assert prop.query_path == expected_query_path
        assert prop.negated is kwargs.get('negated', False)
        assert prop.cached is kwargs.get('cached', RelatedExistenceCheckProperty.cached)

    @pytest.mark.parametrize('path', ['my_field', 'my_relation__my_field'])
    def test_base_condition(self, path):
        prop = RelatedExistenceCheckProperty(path)
        condition = prop._base_condition
        assert isinstance(condition, Q)
        assert len(condition.children) == 1
        assert condition.children[0] == (six.text_type(QueryPath(path) + 'isnull'), False)

    @pytest.mark.parametrize('negated', [False, True])
    def test_getter(self, monkeypatch, categories, applications, negated):
        monkeypatch.setattr(get_queryable_property(CategoryWithClassBasedProperties, 'has_versions'),
                            'negated', negated)
        assert categories[0].has_versions is not negated
        assert categories[1].has_versions is not negated
        applications[1].versions.all().delete()
        assert categories[0].has_versions is not negated
        assert categories[1].has_versions is negated

    @pytest.mark.parametrize('negated', [False, True])
    def test_getter_based_on_non_relation_field(self, monkeypatch, applications, negated):
        monkeypatch.setattr(get_queryable_property(ApplicationWithClassBasedProperties, 'has_version_with_changelog'),
                            'negated', negated)
        assert applications[0].has_version_with_changelog is not negated
        assert applications[1].has_version_with_changelog is not negated
        applications[0].versions.filter(major=2).delete()
        assert applications[0].has_version_with_changelog is negated
        assert applications[1].has_version_with_changelog is not negated

    @pytest.mark.parametrize('negated', [False, True])
    def test_filter(self, monkeypatch, categories, applications, negated):
        monkeypatch.setattr(get_queryable_property(CategoryWithClassBasedProperties, 'has_versions'),
                            'negated', negated)
        queryset = CategoryWithClassBasedProperties.objects.all()
        assert set(queryset.filter(has_versions=not negated)) == set(categories[:2])
        assert not queryset.filter(has_versions=negated).exists()
        applications[1].versions.all().delete()
        assert queryset.get(has_versions=not negated) == categories[0]
        assert queryset.get(has_versions=negated) == categories[1]

    @pytest.mark.parametrize('negated', [False, True])
    def test_filter_based_on_non_relation_field(self, monkeypatch, categories, applications, negated):
        monkeypatch.setattr(get_queryable_property(ApplicationWithClassBasedProperties, 'has_version_with_changelog'),
                            'negated', negated)
        app_queryset = ApplicationWithClassBasedProperties.objects.all()
        category_queryset = CategoryWithClassBasedProperties.objects.all()
        assert set(app_queryset.filter(has_version_with_changelog=not negated)) == set(applications[:2])
        assert not app_queryset.filter(has_version_with_changelog=negated).exists()
        assert (set(category_queryset.filter(applications__has_version_with_changelog=not negated)) ==
                set(categories[:2]))
        assert not category_queryset.filter(applications__has_version_with_changelog=negated).exists()
        applications[1].versions.filter(major=2).delete()
        assert app_queryset.get(has_version_with_changelog=not negated) == applications[0]
        assert app_queryset.get(has_version_with_changelog=negated) == applications[1]
        assert category_queryset.get(applications__has_version_with_changelog=not negated) == categories[0]
        assert category_queryset.get(applications__has_version_with_changelog=negated) == categories[1]

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    def test_annotation(self, categories, applications):
        queryset = CategoryWithClassBasedProperties.objects.all()
        assert list(queryset.order_by('has_versions', 'pk')) == categories[:2]
        applications[1].versions.all().delete()
        assert list(queryset.order_by('has_versions', 'pk')) == [categories[1], categories[0]]

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
    def test_annotation_based_on_non_relation_field(self, categories, applications):
        app_queryset = ApplicationWithClassBasedProperties.objects.all()
        category_queryset = CategoryWithClassBasedProperties.objects.all()
        assert list(app_queryset.order_by('-has_version_with_changelog', '-pk')) == [applications[1], applications[0]]
        assert list(category_queryset.order_by('-applications__has_version_with_changelog', '-pk')) == [
            categories[1], categories[0], categories[0]]
        applications[1].versions.filter(major=2).delete()
        assert list(app_queryset.order_by('-has_version_with_changelog', '-pk')) == applications[:2]
        assert list(category_queryset.order_by('-applications__has_version_with_changelog', '-pk')) == [
            categories[0], categories[1], categories[0]]



# encoding: utf-8

from functools import wraps

import six
from django.db.models import BooleanField

from ..exceptions import QueryablePropertyError
from ..managers import QueryablePropertiesQuerySetMixin
from ..utils.internal import InjectableMixin, QueryPath

REMAINING_LOOKUPS = '*'  #: A Constant that can be used instead of lookup names to match all remaining lookups.


class LookupFilterMeta(type):
    """
    Metaclass for classes that use the :class:`LookupFilterMixin` to detect the
    individual registered filter methods and make them available to the main
    filter method.
    """

    def __new__(mcs, name, bases, attrs):
        # Find all methods that have been marked with lookups via the
        # `lookup_filter` decorator.
        mappings = {}
        for attr_name, attr in six.iteritems(attrs):
            if callable(attr) and hasattr(attr, '_lookups'):
                for lookup in attr._lookups:
                    mappings[lookup] = attr_name

        # Let the class construction take care of the lookup mappings of the
        # base class(es) and add the ones from the current class to them.
        cls = super(LookupFilterMeta, mcs).__new__(mcs, name, bases, attrs)
        cls._lookup_mappings = dict(cls._lookup_mappings, **mappings)
        return cls


class LookupFilterMixin(six.with_metaclass(LookupFilterMeta, InjectableMixin)):
    """
    A mixin for queryable properties that allows to implement queryset
    filtering via individual methods for different lookups.
    """

    # Avoid overriding the __reduce__ implementation of queryable properties.
    _dynamic_pickling = False

    # Stores mappings of lookups to the names of their corresponding filter
    # functions.
    _lookup_mappings = {}

    remaining_lookups_via_parent = False

    def __init__(self, *args, **kwargs):
        self.lookup_mappings = {lookup: getattr(self, name) for lookup, name in six.iteritems(self._lookup_mappings)}
        super(LookupFilterMixin, self).__init__(*args, **kwargs)

    @classmethod
    def lookup_filter(cls, *lookups):
        """
        Decorator for individual filter methods of classes that use the
        :class:`LookupFilterMixin` to register the decorated methods for the
        given lookups.

        :param str lookups: The lookups to register the decorated method for.
        :return: The actual internal decorator.
        :rtype: function
        """
        def decorator(func):
            func._lookups = lookups  # Store the lookups on the function to be able to read them in the meta class.
            return func
        return decorator

    @classmethod
    def boolean_filter(cls, method):
        """
        Decorator for individual filter methods of classes that use the
        :class:`LookupFilterMixin` to register the methods that are simple
        boolean filters (i.e. the filter can only be called with a ``True`` or
        ``False`` value). This automatically restricts the usable lookups to
        ``exact``. Decorated methods should not expect the ``lookup`` and
        ``value`` parameters and should always return a ``Q`` object
        representing the positive (i.e. ``True``) filter case. The decorator
        will automatically negate the condition if the filter was called with a
        ``False`` value.

        :param function method: The method to decorate.
        :return: The decorated method.
        :rtype: function
        """
        @wraps(method)
        def filter_wrapper(self, model, lookup, value):
            """Actual filter method that negates the condition if required."""
            condition = method(self, model)
            if not value:
                condition.negate()
            return condition
        lookup_decorator = cls.lookup_filter('exact')
        return lookup_decorator(filter_wrapper)

    def get_filter(self, cls, lookup, value):
        # Resolve the correct method to call for the given lookup in this order:
        # 1. Check if there is an explicit method for the given lookup.
        # 2. Check if a method is configured for REMAINING_LOOKUPS.
        # 3. Check if a fallback to the parent class implementation is allowed.
        method = self.lookup_mappings.get(lookup) or self.lookup_mappings.get(REMAINING_LOOKUPS)
        if not method:
            if not self.remaining_lookups_via_parent:
                raise QueryablePropertyError(
                    'Queryable property "{prop}" does not implement filtering with lookup "{lookup}".'
                    .format(prop=self, lookup=lookup)
                )
            method = super(LookupFilterMixin, self).get_filter
        return method(cls, lookup, value)


# Aliases to allow the usage of the decorators without the "LookupFilterMixin."
# prefix.
boolean_filter = LookupFilterMixin.boolean_filter
lookup_filter = LookupFilterMixin.lookup_filter


class SetterMixin(object):
    """
    A mixin for queryable properties that also define a setter.
    """

    def set_value(self, obj, value):  # pragma: no cover
        """
        Setter method for the queryable property, which will be called when the
        property is write-accessed.

        :param django.db.models.Model obj: The object on which the property was
                                           accessed.
        :param value: The value to set.
        """
        raise NotImplementedError()


class AnnotationMixin(InjectableMixin):
    """
    A mixin for queryable properties that allows to add an annotation to
    represent them to querysets.
    """

    # Avoid overriding the __reduce__ implementation of queryable properties.
    _dynamic_pickling = False

    filter_requires_annotation = True

    @property
    def admin_order_field(self):
        """
        Return the field name for the ordering in the admin, which is simply
        the property's name since it's annotatable.

        :return: The field name for ordering in the admin.
        :rtype: str
        """
        return self.name

    def get_annotation(self, cls):  # pragma: no cover
        """
        Construct an annotation representing this property that can be added
        to querysets of the model associated with this property.

        :param type cls: The model class of which a queryset should be
                         annotated.
        :return: An annotation object.
        """
        raise NotImplementedError()

    def get_filter(self, cls, lookup, value):
        # Since annotations can be filtered like regular fields, a Q object
        # that simply passes the filter through can be used.
        return (QueryPath(self.name) + lookup).build_filter(value)


class AnnotationGetterMixin(AnnotationMixin):
    """
    A mixin for queryable properties that support annotation and use their
    annotation even to provide the value for their getter (i.e. perform a query
    to retrieve the getter value).
    """

    def __init__(self, cached=None, *args, **kwargs):
        """
        Initialize a new queryable property based that uses the
        :class:`AnnotationGetterMixin`.

        :param cached: Determines if values obtained by the getter should be
                       cached (similar to ``cached_property``). A value of None
                       means using the default value.
        """
        super(AnnotationGetterMixin, self).__init__(*args, **kwargs)
        if cached is not None:
            self.cached = cached

    def get_value(self, obj):
        queryset = self.get_queryset_for_object(obj).distinct().select_properties(self.name)
        return queryset.values_list(self.name, flat=True).get()

    def get_queryset(self, model):
        """
        Construct a base queryset for the given model class that can be used
        to build queries in property code.

        :param model: The model class to build the queryset for.
        """
        # Inject the mixin to be able to use select_properties in the getter.
        return QueryablePropertiesQuerySetMixin.inject_into_object(model._base_manager.all())

    def get_queryset_for_object(self, obj):
        """
        Construct a base queryset that can be used to retrieve the getter value
        for the given object.

        :param django.db.models.Model obj: The object to build the queryset
                                           for.
        :return: A base queryset for the correct model that is already filtered
                 for the given object.
        :rtype: django.db.models.QuerySet
        """
        return self.get_queryset(obj.__class__).filter(pk=obj.pk)


class UpdateMixin(object):
    """
    A mixin for queryable properties that allows to use themselves in update
    queries.
    """

    def get_update_kwargs(self, cls, value):  # pragma: no cover
        """
        Resolve an update keyword argument for this property into the actual
        keyword arguments to emulate an update using this property.

        :param type cls: The model class of which an update query should be
                         performed.
        :param value: The value passed to the update call for this property.
        :return: The actual keyword arguments to set in the update call instead
                 of the given one.
        :rtype: dict
        """
        raise NotImplementedError()


class BooleanMixin(LookupFilterMixin):
    """
    Internal mixin class for common properties that return boolean values,
    which is intended to be used in conjunction with one of the annotation
    mixins.
    """

    filter_requires_annotation = False

    def _get_condition(self, cls):  # pragma: no cover
        """
        Build the query filter condition for this boolean property, which is
        used for both the filter and the annotation implementation.

        :param type cls: The model class of which a queryset should be filtered
                         or annotated.
        :return: The filter condition for this property.
        :rtype: django.db.models.Q
        """
        raise NotImplementedError()

    @boolean_filter
    def get_exact_filter(self, cls):
        return self._get_condition(cls)

    def get_annotation(self, cls):
        from django.db.models import Case, When

        return Case(
            When(self._get_condition(cls), then=True),
            default=False,
            output_field=BooleanField()
        )


class SubqueryMixin(AnnotationGetterMixin):
    """
    Internal mixin class for common properties that are based on custom
    subqueries.
    """

    def __init__(self, queryset, **kwargs):
        """
        Initialize a new subquery-based queryable property.

        :param queryset: The internal queryset to use as the subquery or a
                         callable without arguments that generates the internal
                         queryset.
        :type queryset: django.db.models.QuerySet | function
        """
        self.queryset = queryset
        super(SubqueryMixin, self).__init__(**kwargs)

    def _build_subquery(self, queryset):  # pragma: no cover
        """
        Build the subquery annotation that should be used to represent this
        queryable property in querysets.

        :param django.db.models.QuerySet queryset: The internal queryset to
                                                   base the annotation on.
        :return: The subquery annotation.
        :rtype: django.db.models.expressions.Subquery
        """
        raise NotImplementedError()

    def get_annotation(self, cls):
        queryset = self.queryset
        if callable(queryset):
            queryset = queryset()
        return self._build_subquery(queryset)



# -*- coding: utf-8 -*-

import pytest
from django import VERSION as DJANGO_VERSION
from django.contrib.admin import site
from django.contrib.admin.filters import ChoicesFieldListFilter, FieldListFilter
from django.db.models.query import QuerySet
from mock import patch

from queryable_properties.compat import ADMIN_QUERYSET_METHOD_NAME, nullcontext
from queryable_properties.managers import QueryablePropertiesQuerySetMixin
from queryable_properties.utils import get_queryable_property
from ..app_management.admin import ApplicationAdmin, VersionAdmin, VersionInline
from ..app_management.models import ApplicationWithClassBasedProperties, VersionWithClassBasedProperties
from .test_checks import DummyListFilter


class TestQueryablePropertiesAdminMixin(object):

    @pytest.mark.parametrize('admin_class, model, expected_value', [
        (VersionAdmin, VersionWithClassBasedProperties, ()),
        (ApplicationAdmin, ApplicationWithClassBasedProperties, ('version_count',)),
        (VersionInline, ApplicationWithClassBasedProperties, ('changes_or_default',)),
    ])
    def test_get_list_select_properties(self, rf, admin_class, model, expected_value):
        admin = admin_class(model, site)
        assert admin.get_list_select_properties(rf.get('/')) == expected_value

    @pytest.mark.parametrize('admin_class, model, apply_patch, expected_selected_properties', [
        (VersionAdmin, VersionWithClassBasedProperties, False, ()),
        (VersionAdmin, VersionWithClassBasedProperties, True, ()),
        (ApplicationAdmin, ApplicationWithClassBasedProperties, False,
         (get_queryable_property(ApplicationWithClassBasedProperties, 'version_count'),)),
        (ApplicationAdmin, ApplicationWithClassBasedProperties, True,
         (get_queryable_property(ApplicationWithClassBasedProperties, 'version_count'),)),
    ])
    def test_get_queryset(self, rf, admin_class, model, apply_patch, expected_selected_properties):
        admin = admin_class(model, site)
        qs_patch = nullcontext()
        if apply_patch:
            qs_patch = patch('django.contrib.admin.options.ModelAdmin.{}'.format(ADMIN_QUERYSET_METHOD_NAME),
                             return_value=QuerySet(model))

        with qs_patch:
            queryset = admin.get_queryset(rf.get('/'))
        assert queryset.model is model
        assert isinstance(queryset, QueryablePropertiesQuerySetMixin)
        assert len(queryset.query._queryable_property_annotations) == len(expected_selected_properties)
        for prop in expected_selected_properties:
            assert any(ref.property is prop for ref in queryset.query._queryable_property_annotations)

    @pytest.mark.parametrize('list_filter_item, property_name', [
        ('name', None),
        (DummyListFilter, None),
        ('support_start_date', 'support_start_date'),
        (('support_start_date', ChoicesFieldListFilter), 'support_start_date'),
    ])
    def test_get_list_filter(self, monkeypatch, rf, list_filter_item, property_name):
        monkeypatch.setattr(ApplicationAdmin, 'list_filter', ('common_data', list_filter_item))
        admin = ApplicationAdmin(ApplicationWithClassBasedProperties, site)
        list_filter = admin.list_filter
        if DJANGO_VERSION >= (1, 5):
            list_filter = admin.get_list_filter(rf.get('/'))
        assert list_filter[0] == 'common_data'
        assert (list_filter[1] == list_filter_item) is (not property_name)
        if property_name:
            replacement = list_filter[1]
            assert callable(replacement)
            filter_instance = replacement(rf.get('/'), {}, admin.model, admin)
            assert isinstance(filter_instance, FieldListFilter)
            assert filter_instance.field.name == property_name

    @pytest.mark.skipif(DJANGO_VERSION < (2, 1), reason='Arbitrary search fields were not allowed before Django 2.1')
    @pytest.mark.django_db
    @pytest.mark.parametrize('search_term, expected_count', [
        ('app', 2),
        ('cool', 1),
        ('another', 1),
        ('not-found', 0),
        ('2.0.0', 1),
        ('1.3.1', 1),
        ('1.3', 1),
        ('1.3.0', 0),
        ('1.2.3', 0),
        ('3.4.5', 0),
    ])
    @pytest.mark.usefixtures('versions')
    def test_get_search_results(self, rf, applications, search_term, expected_count):
        applications[0].versions.filter(version='2.0.0').delete()
        request = rf.get('/')
        admin = ApplicationAdmin(ApplicationWithClassBasedProperties, site)
        queryset = admin.get_search_results(request, admin.get_queryset(request), search_term)[0]
        assert queryset.count() == expected_count



# -*- coding: utf-8 -*-

from queryable_properties.admin import QueryablePropertiesAdmin, QueryablePropertiesTabularInline
from ..app_management.models import VersionWithClassBasedProperties


class VersionAdmin(QueryablePropertiesAdmin):
    list_display = ('version', 'application', 'is_supported')
    list_filter = ('application', 'major')
    search_fields = ('changes',)
    date_hierarchy = 'supported_from'
    ordering = ('application', 'major', 'minor', 'patch')


class VersionInline(QueryablePropertiesTabularInline):
    model = VersionWithClassBasedProperties
    list_select_properties = ('changes_or_default',)
    ordering = ('version',)


class ApplicationAdmin(QueryablePropertiesAdmin):
    list_display = ('name', 'highest_version', 'version_count')
    list_filter = ('common_data', 'support_start_date')
    list_select_properties = ('version_count',)
    search_fields = ('name', 'highest_version')
    ordering = ('name', 'version_count')
    inlines = (VersionInline,)



# encoding: utf-8
from collections import Counter

import pytest
import six
from django import VERSION as DJANGO_VERSION
from django.db.models import CharField, Count, IntegerField, Q, Sum
from six.moves import cPickle

from queryable_properties.exceptions import QueryablePropertyDoesNotExist, QueryablePropertyError
from queryable_properties.properties.base import QueryablePropertyDescriptor
from queryable_properties.utils import get_queryable_property
from queryable_properties.utils.internal import (
    MISSING_OBJECT, InjectableMixin, ModelAttributeGetter, NodeChecker, NodeModifier, NodeProcessor,
    QueryablePropertyReference, QueryPath, get_output_field, get_queryable_property_descriptor,
    parametrizable_decorator, resolve_queryable_property,
)
from ..app_management.models import (
    ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties, CategoryWithClassBasedProperties,
    CategoryWithDecoratorBasedProperties, VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties,
)
from ..conftest import Concat, Value


class BaseMetaclass(type):
    pass


class MixinMetaclass(type):
    pass


class DummyClass(object):

    def __init__(self, attr1, attr2):
        self.attr1 = attr1
        self.attr2 = attr2


class DummyMixin(InjectableMixin):

    def init_injected_attrs(self):
        self.mixin_attr1 = 1.337
        self.mixin_attr2 = 'test'


class DummyClassWithMetaclass(six.with_metaclass(BaseMetaclass)):
    pass


class DummyMixinWithMetaclass(six.with_metaclass(MixinMetaclass, InjectableMixin)):
    pass


@parametrizable_decorator
def decorator(function, *args, **kwargs):
    function.args = args
    function.kwargs = kwargs
    return function


class TestQueryPath(object):

    @pytest.mark.parametrize('path, expected_result', [
        ([], QueryPath()),
        (('a', 'b'), QueryPath(('a', 'b'))),
        ('a__b', QueryPath(('a', 'b'))),
    ])
    def test_constructor(self, path, expected_result):
        query_path = QueryPath(path)
        assert query_path == expected_result

    @pytest.mark.parametrize('query_path, addition, expected_result', [
        (QueryPath(), QueryPath(['a']), QueryPath(('a',))),
        (QueryPath('a'), ('b', 'c'), QueryPath(('a', 'b', 'c'))),
        (QueryPath('a'), ['b'], QueryPath(('a', 'b'))),
        (QueryPath(('a', 'b')), 'c__d', QueryPath(('a', 'b', 'c', 'd'))),
    ])
    def test_add(self, query_path, addition, expected_result):
        result = query_path + addition
        assert isinstance(result, QueryPath)
        assert result == expected_result

    @pytest.mark.parametrize('query_path, item, expected_result', [
        (QueryPath('a__b'), 0, 'a'),
        (QueryPath('a__b'), slice(0, 1), QueryPath('a')),
        (QueryPath('a__b'), slice(5, 10), QueryPath()),
    ])
    def test_get_item(self, query_path, item, expected_result):
        result = query_path[item]
        assert isinstance(result, expected_result.__class__)
        assert result == expected_result

    def test_string_representation(self):
        query_path = QueryPath(('a', 'b', 'c'))
        assert six.text_type(query_path) == 'a__b__c'

    def test_representation(self):
        query_path = QueryPath(('a', 'b'))
        assert repr(query_path) == '<QueryPath: a__b>'

    def test_build_filter(self):
        path = 'a__b__c'
        value = 1337
        query_path = QueryPath(path)
        condition = query_path.build_filter(value)
        assert isinstance(condition, Q)
        assert len(condition.children) == 1
        assert condition.children[0] == (path, value)


class TestInjectableMixin(object):

    @pytest.mark.parametrize('class_name, expected_class_name', [
        (None, DummyClass.__name__),
        ('TestClass', 'TestClass'),
    ])
    def test_mix_with_class(self, monkeypatch, class_name, expected_class_name):
        monkeypatch.setattr(DummyMixin, '_created_classes', {})
        assert not DummyMixin._created_classes
        created_classes = set()

        # Execute the code twice to test the cache
        for _ in range(2):
            cls = DummyMixin.mix_with_class(DummyClass, class_name)
            # Test that trying to mix in the class a second time returns the
            # base class unchanged.
            assert cls is DummyMixin.mix_with_class(cls)
            created_classes.add(cls)
            assert issubclass(cls, DummyClass)
            assert issubclass(cls, DummyMixin)
            assert cls.__name__ == expected_class_name
            assert len(DummyMixin._created_classes) == 1
            assert len(created_classes) == 1
            # Test that the __init__ method of the new class correctly
            # initializes the injected attributes.
            obj = cls(5, 'abc')
            assert obj.mixin_attr1 == 1.337
            assert obj.mixin_attr2 == 'test'

    @pytest.mark.parametrize('base_class, mixin_class', [
        (DummyClass, DummyMixin),
        (DummyClass, DummyMixinWithMetaclass),
        (DummyClassWithMetaclass, DummyMixin),
        (DummyClassWithMetaclass, DummyMixinWithMetaclass),
    ])
    def test_mix_with_class_metaclasses(self, base_class, mixin_class):
        cls = mixin_class.mix_with_class(base_class)
        assert issubclass(cls, base_class)
        assert issubclass(cls, mixin_class)
        assert isinstance(cls, base_class.__class__)
        assert isinstance(cls, mixin_class.__class__)

    @pytest.mark.parametrize('init', [True, False])
    def test_inject_into_object(self, init):
        obj = DummyClass(5, 'abc')
        obj.mixin_attr1 = obj.mixin_attr2 = None
        assert DummyMixin.inject_into_object(obj, init=init) is obj
        assert isinstance(obj, DummyClass)
        assert isinstance(obj, DummyMixin)
        assert obj.attr1 == 5
        assert obj.attr2 == 'abc'
        assert obj.mixin_attr1 == (1.337 if init else None)
        assert obj.mixin_attr2 == ('test' if init else None)
        # Test that init_injected_attrs is not called when no injection takes
        # places due to the object already using the mixin.
        obj.mixin_attr1 = obj.mixin_attr2 = None
        DummyMixin.inject_into_object(obj)
        assert obj.mixin_attr1 is obj.mixin_attr2 is None

    def test_pickle_unpickle(self):
        base_obj = DummyClass('xyz', 42.42)
        DummyMixin.inject_into_object(base_obj)
        serialized_obj = cPickle.dumps(base_obj)
        deserialized_obj = cPickle.loads(serialized_obj)

        for obj in (base_obj, deserialized_obj):
            assert isinstance(obj, DummyClass)
            assert isinstance(obj, DummyMixin)
            assert obj.attr1 == 'xyz'
            assert obj.attr2 == 42.42
            assert obj.mixin_attr1 == 1.337
            assert obj.mixin_attr2 == 'test'

    def test_no_reduce_implementation(self, monkeypatch):
        monkeypatch.setattr(DummyMixin, '_dynamic_pickling', False)
        monkeypatch.setattr(DummyMixin, '_created_classes', {})
        base_obj = DummyClass('xyz', 42.42)
        DummyMixin.inject_into_object(base_obj)
        with pytest.raises(cPickle.PicklingError):
            cPickle.dumps(base_obj)


class TestNodeProcessor(object):

    def test_initializer(self):
        def func(item):
            return item

        processor = NodeProcessor(func)
        assert processor.func is func

    def test_iter_leaves(self):
        processor = NodeProcessor(lambda item: item)
        inner_q = Q(a=1) | Q(b=2)
        outer_q = Q(inner_q, c=3)
        assert list(processor.iter_leaves(outer_q)) == [
            (inner_q, 0, ('a', 1)),
            (inner_q, 1, ('b', 2)),
            (outer_q, 1, ('c', 3)),
        ]


class TestNodeChecker(object):

    @pytest.mark.parametrize('node, path, expected_result', [
        (Q(a=1), 'a', True),
        (Q(b=2), 'a', False),
        (Q(a=1), 'b', False),
        (Q(b=2), 'b', True),
        (Q(Q(a=1) | Q(b=2), c=3), 'a', True),
        (Q(Q(a=1) | Q(b=2), c=3), 'c', True),
        (Q(Q(a=1) | Q(b=2), c=3), 'd', False),
    ])
    def test_check_leaves(self, node, path, expected_result):
        # The predicate checks if a leaf for a given path exists
        checker = NodeChecker(lambda item, required_path: item[0] == path)
        assert checker.check_leaves(node, required_path=path) is expected_result


class TestNodeModifier(object):

    @pytest.mark.parametrize('copy', [True, False])
    @pytest.mark.parametrize('increment, expected_a, expected_b, expected_c', [
        (1, 2, 3, 4),
        (10, 11, 12, 13),
    ])
    def test_modify_leaves(self, copy, increment, expected_a, expected_b, expected_c):
        modifier = NodeModifier(lambda item, inc: ('new_{}'.format(item[0]), item[1] + inc))
        q = Q(Q(a=1) | Q(b=2), c=3)
        result = modifier.modify_leaves(q, copy, inc=increment)
        assert (result is q) is not copy
        assert len(result.children) == 2
        assert result.children[0].children == [('new_a', expected_a), ('new_b', expected_b)]
        assert result.children[1] == ('new_c', expected_c)


class TestModelAttributeGetter(object):

    @pytest.mark.parametrize('path, expected_query_path', [
        ('attr', QueryPath('attr')),
        ('attr1.attr2', QueryPath('attr1__attr2')),
        ('attr1.attr2.attr3', QueryPath('attr1__attr2__attr3')),
        (['attr1', 'attr2'], QueryPath('attr1__attr2')),
        (('attr1', 'attr2', 'attr3'), QueryPath('attr1__attr2__attr3')),
    ])
    def test_initializer(self, path, expected_query_path):
        getter = ModelAttributeGetter(path)
        assert getter.query_path == expected_query_path

    @pytest.mark.django_db
    @pytest.mark.parametrize('attribute_name, expected_value', [
        ('major', 1),
        ('changes', None),
    ])
    def test_get_attribute(self, versions, attribute_name, expected_value):
        getter = ModelAttributeGetter(())
        assert getter._get_attribute(versions[0], attribute_name) == expected_value

    def test_get_attribute_catch_attribute_error_on_none(self):
        getter = ModelAttributeGetter(())
        assert getter._get_attribute(None, 'non_existent') is MISSING_OBJECT

    @pytest.mark.django_db
    def test_get_attribute_bubble_attribute_error(self, versions):
        getter = ModelAttributeGetter(())
        with pytest.raises(AttributeError):
            getter._get_attribute(versions[0], 'non_existent')

    @pytest.mark.django_db
    def test_get_attribute_catch_object_does_not_exist(self, applications, versions):
        obj = versions[0].__class__.objects.get(pk=versions[0].pk)  # Refresh from DB
        applications[0].delete()
        getter = ModelAttributeGetter(())
        assert getter._get_attribute(obj, 'application') is MISSING_OBJECT

    @pytest.mark.django_db
    @pytest.mark.parametrize('path, expected_value', [
        ('major', 1),
        ('changes', None),
        ('application.name', 'My cool App'),
    ])
    def test_get_value(self, versions, path, expected_value):
        obj = versions[0]
        getter = ModelAttributeGetter(path)
        assert getter.get_value(obj) == expected_value

    @pytest.mark.django_db
    def test_get_value_missing_object(self, applications, versions):
        obj = versions[0].__class__.objects.get(pk=versions[0].pk)  # Refresh from DB
        applications[0].delete()
        getter = ModelAttributeGetter('application.name')
        assert getter.get_value(obj) is MISSING_OBJECT

    @pytest.mark.django_db
    @pytest.mark.usefixtures('versions')
    def test_get_values(self, categories):
        getter = ModelAttributeGetter('applications.versions.major')
        assert Counter(getter.get_values(categories[0])) == Counter({2: 2, 1: 6})

    @pytest.mark.django_db
    @pytest.mark.usefixtures('versions')
    def test_get_values_missing_object(self, categories):
        getter = ModelAttributeGetter('applications.versions.major')
        VersionWithClassBasedProperties.objects.all().delete()
        assert getter.get_values(categories[0]) == []
        ApplicationWithClassBasedProperties.objects.all().delete()
        assert getter.get_values(categories[0]) == []

    @pytest.mark.parametrize('path, lookup, value, expected_query_name', [
        ('attr', 'exact', 1337, 'attr__exact'),
        ('attr1.attr2', 'in', ('test', 'value'), 'attr1__attr2__in'),
        ('attr1.attr2.attr3', 'gt', 1337, 'attr1__attr2__attr3__gt'),
    ])
    def test_build_filter(self, path, lookup, value, expected_query_name):
        getter = ModelAttributeGetter(path)
        condition = getter.build_filter(lookup, value)
        assert isinstance(condition, Q)
        assert len(condition.children) == 1
        assert condition.children[0] == (expected_query_name, value)


def test_parametrizable_decorator():

    @decorator
    def func1():
        pass

    @decorator(some_kwarg=1, another_kwarg='test')
    def func2():
        pass

    def func3():
        pass

    func3 = decorator(func3, 1, 2, kwarg='test')

    assert func1.args == ()
    assert func1.kwargs == {}
    assert func2.args == ()
    assert func2.kwargs == dict(some_kwarg=1, another_kwarg='test')
    assert func3.args == (1, 2)
    assert func3.kwargs == dict(kwarg='test')


class TestGetQueryablePropertyDescriptor(object):

    @pytest.mark.parametrize('model, property_name', [
        (VersionWithClassBasedProperties, 'major_minor'),
        (VersionWithDecoratorBasedProperties, 'major_minor'),
        (VersionWithClassBasedProperties, 'version'),
        (VersionWithDecoratorBasedProperties, 'version'),
    ])
    def test_property_found(self, model, property_name):
        descriptor = get_queryable_property_descriptor(model, property_name)
        assert isinstance(descriptor, QueryablePropertyDescriptor)

    @pytest.mark.parametrize('model, property_name', [
        (VersionWithClassBasedProperties, 'non_existent'),
        (VersionWithDecoratorBasedProperties, 'non_existent'),
        (VersionWithClassBasedProperties, 'major'),  # Existing model field
        (VersionWithDecoratorBasedProperties, 'major'),  # Existing model field
    ])
    def test_exception(self, model, property_name):
        with pytest.raises(QueryablePropertyDoesNotExist):
            get_queryable_property_descriptor(model, property_name)


class TestQueryablePropertyReference(object):

    @pytest.mark.parametrize('relation_path, expected_result', [
        (QueryPath(), QueryPath('dummy')),
        (QueryPath('application'), QueryPath('application__dummy')),
    ])
    def test_full_path(self, relation_path, expected_result):
        prop = get_queryable_property(ApplicationWithClassBasedProperties, 'dummy')
        ref = QueryablePropertyReference(prop, prop.model, relation_path)
        assert ref.full_path == expected_result

    def test_descriptor(self):
        prop = get_queryable_property(ApplicationWithClassBasedProperties, 'dummy')
        ref = QueryablePropertyReference(prop, prop.model, QueryPath())
        assert ref.descriptor == ApplicationWithClassBasedProperties.dummy

    @pytest.mark.parametrize('lookups, relation_path, expected_filter', [
        (QueryPath(), QueryPath(), 'version_count__exact'),
        (QueryPath('lt'), QueryPath(), 'version_count__lt'),
        (QueryPath('date__year'), QueryPath('application'), 'application__version_count__date__year'),
    ])
    def test_get_filter(self, lookups, relation_path, expected_filter):
        prop = get_queryable_property(ApplicationWithClassBasedProperties, 'version_count')
        ref = QueryablePropertyReference(prop, prop.model, relation_path)
        q = ref.get_filter(lookups, 1337)
        assert isinstance(q, Q)
        assert q.children == [(expected_filter, 1337)]

    def test_get_filter_exception(self):
        prop = get_queryable_property(ApplicationWithClassBasedProperties, 'dummy')
        ref = QueryablePropertyReference(prop, prop.model, QueryPath())
        with pytest.raises(QueryablePropertyError):
            ref.get_filter(QueryPath(), None)

    def test_get_annotation(self):
        prop = get_queryable_property(ApplicationWithClassBasedProperties, 'version_count')
        ref = QueryablePropertyReference(prop, prop.model, QueryPath())
        assert isinstance(ref.get_annotation(), Count)

    def test_get_annotation_exception(self):
        prop = get_queryable_property(ApplicationWithClassBasedProperties, 'dummy')
        ref = QueryablePropertyReference(prop, prop.model, QueryPath())
        with pytest.raises(QueryablePropertyError):
            ref.get_annotation()


class TestResolveQueryableProperty(object):

    @pytest.mark.parametrize('model, query_path, expected_property, expected_lookups', [
        # No relation involved
        (VersionWithClassBasedProperties, QueryPath('version'),
         get_queryable_property(VersionWithClassBasedProperties, 'version'), QueryPath()),
        (VersionWithDecoratorBasedProperties, QueryPath('version'),
         get_queryable_property(VersionWithDecoratorBasedProperties, 'version'), QueryPath()),
        (VersionWithClassBasedProperties, QueryPath('version__lower__exact'),
         get_queryable_property(VersionWithClassBasedProperties, 'version'), QueryPath('lower__exact')),
        (VersionWithDecoratorBasedProperties, QueryPath('version__lower__exact'),
         get_queryable_property(VersionWithDecoratorBasedProperties, 'version'), QueryPath('lower__exact')),
        # FK forward relation
        (VersionWithClassBasedProperties, QueryPath('application__version_count'),
         get_queryable_property(ApplicationWithClassBasedProperties, 'version_count'), QueryPath()),
        (VersionWithDecoratorBasedProperties, QueryPath('application__version_count'),
         get_queryable_property(ApplicationWithDecoratorBasedProperties, 'version_count'), QueryPath()),
        (VersionWithClassBasedProperties, QueryPath('application__major_sum__gt'),
         get_queryable_property(ApplicationWithClassBasedProperties, 'major_sum'), QueryPath('gt')),
        (VersionWithDecoratorBasedProperties, QueryPath('application__major_sum__gt'),
         get_queryable_property(ApplicationWithDecoratorBasedProperties, 'major_sum'), QueryPath('gt')),
        # FK reverse relation
        (ApplicationWithClassBasedProperties, QueryPath('versions__major_minor'),
         get_queryable_property(VersionWithClassBasedProperties, 'major_minor'), QueryPath()),
        (ApplicationWithDecoratorBasedProperties, QueryPath('versions__major_minor'),
         get_queryable_property(VersionWithDecoratorBasedProperties, 'major_minor'), QueryPath()),
        (ApplicationWithClassBasedProperties, QueryPath('versions__version__lower__contains'),
         get_queryable_property(VersionWithClassBasedProperties, 'version'), QueryPath('lower__contains')),
        (ApplicationWithDecoratorBasedProperties, QueryPath('versions__version__lower__contains'),
         get_queryable_property(VersionWithDecoratorBasedProperties, 'version'), QueryPath('lower__contains')),
        # M2M forward relation
        (ApplicationWithClassBasedProperties, QueryPath('categories__circular'),
         get_queryable_property(CategoryWithClassBasedProperties, 'circular'), QueryPath()),
        (ApplicationWithDecoratorBasedProperties, QueryPath('categories__circular'),
         get_queryable_property(CategoryWithDecoratorBasedProperties, 'circular'), QueryPath()),
        (ApplicationWithClassBasedProperties, QueryPath('categories__circular__exact'),
         get_queryable_property(CategoryWithClassBasedProperties, 'circular'), QueryPath('exact')),
        (ApplicationWithDecoratorBasedProperties, QueryPath('categories__circular__exact'),
         get_queryable_property(CategoryWithDecoratorBasedProperties, 'circular'), QueryPath('exact')),
        # M2M reverse relation
        (CategoryWithClassBasedProperties, QueryPath('applications__major_sum'),
         get_queryable_property(ApplicationWithClassBasedProperties, 'major_sum'), QueryPath()),
        (CategoryWithDecoratorBasedProperties, QueryPath('applications__major_sum'),
         get_queryable_property(ApplicationWithDecoratorBasedProperties, 'major_sum'), QueryPath()),
        (CategoryWithClassBasedProperties, QueryPath('applications__version_count__lt'),
         get_queryable_property(ApplicationWithClassBasedProperties, 'version_count'), QueryPath('lt')),
        (CategoryWithDecoratorBasedProperties, QueryPath('applications__version_count__lt'),
         get_queryable_property(ApplicationWithDecoratorBasedProperties, 'version_count'), QueryPath('lt')),
        # Multiple relations
        (CategoryWithClassBasedProperties, QueryPath('applications__versions__application__categories__circular'),
         get_queryable_property(CategoryWithClassBasedProperties, 'circular'), QueryPath()),
        (CategoryWithDecoratorBasedProperties, QueryPath('applications__versions__application__categories__circular'),
         get_queryable_property(CategoryWithDecoratorBasedProperties, 'circular'), QueryPath()),
        (VersionWithClassBasedProperties, QueryPath('application__categories__circular__in'),
         get_queryable_property(CategoryWithClassBasedProperties, 'circular'), QueryPath('in')),
        (VersionWithDecoratorBasedProperties, QueryPath('application__categories__circular__in'),
         get_queryable_property(CategoryWithDecoratorBasedProperties, 'circular'), QueryPath('in')),
    ])
    def test_successful(self, model, query_path, expected_property, expected_lookups):
        expected_ref = QueryablePropertyReference(expected_property, expected_property.model,
                                                  query_path[:-len(expected_lookups) - 1])
        assert resolve_queryable_property(model, query_path) == (expected_ref, expected_lookups)

    @pytest.mark.parametrize('model, query_path', [
        # No relation involved
        (VersionWithClassBasedProperties, QueryPath('non_existent')),
        (VersionWithDecoratorBasedProperties, QueryPath('non_existent')),
        (VersionWithClassBasedProperties, QueryPath('major')),
        (VersionWithDecoratorBasedProperties, QueryPath('major')),
        # FK forward relation
        (VersionWithClassBasedProperties, QueryPath('application__non_existent__exact')),
        (VersionWithDecoratorBasedProperties, QueryPath('application__non_existent__exact')),
        (VersionWithClassBasedProperties, QueryPath('application__name')),
        (VersionWithDecoratorBasedProperties, QueryPath('application__name')),
        # FK reverse relation
        (ApplicationWithClassBasedProperties, QueryPath('versions__non_existent')),
        (ApplicationWithDecoratorBasedProperties, QueryPath('versions__non_existent')),
        (ApplicationWithClassBasedProperties, QueryPath('versions__minor__gt')),
        (ApplicationWithDecoratorBasedProperties, QueryPath('versions__minor__gt')),
        # M2M forward relation
        (ApplicationWithClassBasedProperties, QueryPath('categories__non_existent')),
        (ApplicationWithDecoratorBasedProperties, QueryPath('categories__non_existent')),
        (ApplicationWithClassBasedProperties, QueryPath('categories__name')),
        (ApplicationWithDecoratorBasedProperties, QueryPath('categories__name')),
        # M2M reverse relation
        (CategoryWithClassBasedProperties, QueryPath('applications__non_existent')),
        (CategoryWithDecoratorBasedProperties, QueryPath('applications__non_existent')),
        (CategoryWithClassBasedProperties, QueryPath('applications__name')),
        (CategoryWithDecoratorBasedProperties, QueryPath('applications__name')),
        # Non existent relation
        (VersionWithClassBasedProperties, QueryPath('non_existent_relation__non_existent__in')),
        (VersionWithDecoratorBasedProperties, QueryPath('non_existent_relation__non_existent__in')),
    ])
    def test_unsuccessful(self, model, query_path):
        assert resolve_queryable_property(model, query_path) == (None, QueryPath())


class TestGetOutputField(object):

    CHAR_FIELD = CharField()
    INTEGER_FIELD = IntegerField(null=True)

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Output fields couldn't be declared before Django 1.8")
    @pytest.mark.parametrize('annotation, expected_result', [
        (Concat(Value('test'), 'some_field', output_field=CHAR_FIELD), CHAR_FIELD),
        (Sum('aggregate', output_field=INTEGER_FIELD), INTEGER_FIELD),
    ])
    def test_success(self, annotation, expected_result):
        assert get_output_field(annotation) is expected_result

    @pytest.mark.parametrize('annotation', [Concat(Value('test'), 'some_field'), Sum('aggregate')])
    def test_none(self, annotation):
        assert get_output_field(annotation) is None



# -*- coding: utf-8 -*-
"""
Internal utilities used by the queryable properties library, which may change
without notice or be removed without deprecation.
"""

from collections import namedtuple
from copy import deepcopy
from functools import wraps

import six
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Manager, Q
from django.utils.decorators import method_decorator
from django.utils.tree import Node

from ..compat import LOOKUP_SEP, get_related_model
from ..exceptions import FieldDoesNotExist, QueryablePropertyDoesNotExist, QueryablePropertyError

MISSING_OBJECT = object()  #: Arbitrary object to represent that an object in an attribute chain is missing.


@six.python_2_unicode_compatible
class QueryPath(tuple):
    """
    A utility class to represent query paths, i.e. paths using Django's
    LOOKUP_SEP as their separator.

    Objects can be used to build the string representation of a query path and
    to combine paths using the ``+`` operator.
    """
    __slots__ = ()

    def __new__(cls, path=()):
        """
        Build a new query path instance using the given path, which may be
        either a string that will be split up using the LOOKUP_SEP or another
        type of iterable that already contains the individual path parts.

        :param collections.Iterable path: The query path to represent as string
                                          or other iterable.
        """
        if isinstance(path, six.string_types):
            path = path.split(LOOKUP_SEP)
        return super(QueryPath, cls).__new__(cls, path)

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            other = self.__class__(other)
        return self.__class__(tuple(self) + tuple(other))

    def __getitem__(self, item):
        result = super(QueryPath, self).__getitem__(item)
        if isinstance(item, slice):
            result = self.__class__(result)
        return result

    def __getslice__(self, i, j):  # pragma: no cover
        return self.__class__(super(QueryPath, self).__getslice__(i, j))

    def __str__(self):
        return LOOKUP_SEP.join(self)

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, six.text_type(self))

    def build_filter(self, value):
        """
        Build a filter condition based on this query path and the given value.

        :param value: The value to filter against.
        :return: The filter condition as a Q object.
        :rtype: django.db.models.Q
        """
        return Q(**{six.text_type(self): value})


class NodeProcessor(object):
    """
    Base class for utilities that work with Django's tree nodes.
    """

    def __init__(self, func):
        """
        Initialize a new node processor.

        :param function func: The function that is applied to nodes and/or
                              their children. The function must take the node
                              item as its first argument as well as potentially
                              more context arguments based on the concrete
                              implementation, which also defines the expected
                              return value.
        """
        self.func = func

    def iter_leaves(self, node):
        """
        Iterate over all leaves of the given node, regardless of the depth of
        potential sub-nodes.

        :param Node node: The node to get the leaves from.
        :return: A generator yielding 3-tuples for each leaf consisting of the
                 (sub-)node the leaf belongs to, the index of the item inside
                 of that leaf and the leaf item itself.
        :rtype: collections.Iterable[Node, int, object]
        """
        for index, child in enumerate(node.children):
            if isinstance(child, Node):
                for result in self.iter_leaves(child):
                    yield result
            else:
                yield node, index, child


class NodeChecker(NodeProcessor):
    """
    A utility to test tree nodes against a condition specified by the
    configured function, which therefore must return a boolean value indicating
    whether or not the condition was met.
    """

    def check_leaves(self, node, **context):
        """
        Check the leaves of the given node object against the configured
        function.

        :param Node node: The node whose leaves should be checked.
        :param context: Additional context parameters that will be passed
                        through to the configured function.
        :return: True if the condition matches at least one leaf; otherwise
                 False.
        :rtype: bool
        """
        for branch_node, index, leaf in self.iter_leaves(node):
            if self.func(leaf, **context):
                return True
        return False


class NodeModifier(NodeProcessor):
    """
    A utility to modify the tree nodes using the configured function, which
    therefore must return the replacement value for a given item.
    """

    def modify_leaves(self, node, copy=True, **context):
        """
        Modify the leaves of the given node object using the configured
        function.

        :param Node node: The node whose leaves should be modified.
        :param bool copy: If True, a copy of the given node will be created and
                          modified, leaving the original node untouched. If
                          False, the original node will be modified in place.
        :param context: Additional context parameters that will be passed
                        through to the configured function.
        :return: The modified node.
        :rtype: Node
        """
        if copy:
            node = deepcopy(node)
        for branch_node, index, leaf in self.iter_leaves(node):
            branch_node.children[index] = self.func(leaf, **context)
        return node


class QueryablePropertyReference(namedtuple('QueryablePropertyReference', 'property model relation_path')):
    """
    A reference to a queryable property that also holds the path to reach the
    property across relations.
    """
    __slots__ = ()
    node_modifier = NodeModifier(lambda item, ref: (six.text_type(ref.relation_path + item[0]), item[1]))

    @property
    def full_path(self):
        """
        Return the full query path to the queryable property (including the
        relation prefix).

        :return: The full path to the queryable property.
        :rtype: QueryPath
        """
        return self.relation_path + self.property.name

    @property
    def descriptor(self):
        """
        Return the descriptor object associated with the queryable property
        this reference points to.

        :return: The queryable property descriptor for the referenced property.
        :rtype: queryable_properties.properties.base.QueryablePropertyDescriptor
        """
        return get_queryable_property_descriptor(self.model, self.property.name)

    def get_filter(self, lookups, value):
        """
        A wrapper for the get_filter method of the property this reference
        points to. It checks if the property actually supports filtering and
        applies the relation path (if any) to the returned Q object.

        :param QueryPath lookups: The lookups/transforms to use for the filter.
        :param value: The value passed to the filter condition.
        :return: A Q object to filter using this property.
        :rtype: django.db.models.Q
        """
        if not self.property.get_filter:
            raise QueryablePropertyError('Queryable property "{}" is supposed to be used as a filter but does not '
                                         'implement filtering.'.format(self.property))

        # Use the model stored on this reference instead of the one on the
        # property since the query may be happening from a subclass of the
        # model the property is defined on.
        q_obj = self.property.get_filter(self.model, six.text_type(lookups) or 'exact', value)
        if self.relation_path:
            # If the resolved property belongs to a related model, all actual
            # conditions in the returned Q object must be modified to use the
            # current relation path as prefix.
            q_obj = self.node_modifier.modify_leaves(q_obj, ref=self)
        return q_obj

    def get_annotation(self):
        """
        A wrapper for the get_annotation method of the property this reference
        points to. It checks if the property actually supports annotation
        creation performs the internal call with the correct model class.

        :return: An annotation object.
        """
        if not self.property.get_annotation:
            raise QueryablePropertyError('Queryable property "{}" needs to be added as annotation but does not '
                                         'implement annotation creation.'.format(self.property))
        # Use the model stored on this reference instead of the one on the
        # property since the query may be happening from a subclass of the
        # model the property is defined on.
        return self.property.get_annotation(self.model)


class InjectableMixin(object):
    """
    A base class for mixin classes that are used to dynamically created classes
    based on a base class and the mixin class.
    """

    # Intentionally use a single cache for all subclasses since it is in no way
    # harmful to use a shared cache.
    _created_classes = {}
    # Class attribute to determine if dynamically built classes should receive
    # a custom __reduce__ implementation so their objects can be pickled.
    _dynamic_pickling = True

    def __init__(self, *args, **kwargs):
        super(InjectableMixin, self).__init__(*args, **kwargs)
        self.init_injected_attrs()

    def init_injected_attrs(self):
        """
        Initialize the attributes this mixin contributes. This method will be
        called during :meth:`__init__` and after the mixin was injected into an
        object.
        """
        pass

    @classmethod
    def mix_with_class(cls, base_class, class_name=None):
        """
        Create a new class based on the given base class and this mixin class.
        If this mixin class is already part of the class hierarchy of the given
        base class, the base class will be returned unchanged.

        :param type base_class: The base class to mix the mixin into.
        :param str class_name: An optional name for the dynamically created
                               class. If None is supplied (default), the class
                               name of the dynamically created class will be
                               the one of the object's original class. Will
                               be applied if a new class is created.
        :return: The generated class or the base class if it already uses this
                 mixin.
        :rtype: type
        """
        if issubclass(base_class, cls):
            return base_class

        class_name = str(class_name or base_class.__name__)
        cache_key = (base_class, cls, class_name)
        created_class = cls._created_classes.get(cache_key)
        if created_class is None:
            attrs = {}
            metaclass = type
            if (not issubclass(cls.__class__, base_class.__class__) and
                    not issubclass(base_class.__class__, cls.__class__)):
                # If the metaclasses of both classes are unrelated, try to build
                # a new metaclass based on both dynamically.
                metaclass = type(base_class.__class__.__name__, (cls.__class__, base_class.__class__), {})
            if cls._dynamic_pickling:
                # Make sure objects of a dynamically created class can be pickled.
                def __reduce__(self):
                    get_state = getattr(self, '__getstate__', lambda: self.__dict__)
                    return _unpickle_injected_object, (base_class, cls, class_name), get_state()
                attrs['__reduce__'] = __reduce__

            created_class = cls._created_classes[cache_key] = metaclass(class_name, (cls, base_class), attrs)
        return created_class

    @classmethod
    def inject_into_object(cls, obj, class_name=None, init=True):
        """
        Update the given object's class by dynamically generating a new class
        based on the object's original class and this mixin class and changing
        the given object into an object of this new class. If this mixin is
        already part of the object's class hierarchy, its class will not
        change.

        :param obj: The object whose class should be changed.
        :param str class_name: An optional name for the dynamically created
                               class. If None is supplied (default), the class
                               name of the dynamically created class will be
                               the one of the object's original class. Will
                               be applied if a new class is created.
        :param bool init: Whether or not to perform the initialization of
                          injected attributes if the object's class was
                          changed.
        :return: The (potentially) modified object.
        """
        new_class = cls.mix_with_class(obj.__class__, class_name)
        if new_class is not obj.__class__:
            obj.__class__ = new_class
            if init:
                obj.init_injected_attrs()
        return obj


# This must be a standalone function for Python 2, where it could not be
# pickled being a static method on the InjectableMixin, even if the underlying
# function had the __safe_for_unpickling__ flag.
def _unpickle_injected_object(base_class, mixin_class, class_name=None):
    """
    Callable for the pickler to unpickle objects of a dynamically created class
    based on the InjectableMixin. It creates the base object from the original
    base class and re-injects the mixin class when unpickling an object.

    :param type base_class: The base class of the pickled object before adding
                            the mixin via injection.
    :param type mixin_class: The :class:`InjectableMixin` subclass that was
                             injected into the pickled object.
    :param str class_name: The class name of the pickled object's dynamically
                           created class.
    :return: The initial unpickled object (before the pickler restores the
             object's state).
    """
    obj = base_class.__new__(base_class, ())
    return mixin_class.inject_into_object(obj, class_name, init=False)


_unpickle_injected_object.__safe_for_unpickling__ = True


class ModelAttributeGetter(object):
    """
    An attribute getter akin to :func:`operator.attrgetter` specifically
    designed for model objects. Like Python's attrgetter, it allows to access
    attributes on related objects using dot-notation, but it catches some
    expected exceptions related to models when following attribute chains.
    It also allows to build filters for querysets based on the configured
    attribute path.
    """

    ATTRIBUTE_SEPARATOR = '.'

    def __init__(self, attribute_path):
        """
        Initialize a new model attribute getter using the specified attribute
        path.

        :param attribute_path: The path to the attribute to retrieve in as
                               an iterable containing the individual parts or
                               a string using dot-notation (see the docs for
                               :func:`operator.attrgetter` for examples). For
                               queryset-related operations, all parts will be
                               combined using the lookup separator (``__``).
        :type attribute_path: collections.Iterable
        """
        if isinstance(attribute_path, six.string_types):
            attribute_path = attribute_path.split(self.ATTRIBUTE_SEPARATOR)
        self.query_path = QueryPath(attribute_path)

    def _get_attribute(self, obj, attribute_name):
        """
        Get and return the value for a single non-nested attribute from the
        given object, catching certain exceptions and returning the
        ``MISSING_OBJECT`` constant in such cases.

        :param obj: The object to get the attribute value from.
        :param str attribute_name: The name of the attribute.
        :return: The attribute value or the ``MISSING_OBJECT`` constant.
        """
        try:
            return getattr(obj, attribute_name)
        except ObjectDoesNotExist:
            # Allow missing DB objects without raising an error, e.g. for
            # reverse one-to-one relations.
            return MISSING_OBJECT
        except AttributeError:
            # Allow objects in between to be None without raising an error,
            # e.g. for nullable fields.
            if obj is None:
                return MISSING_OBJECT
            raise

    def get_value(self, obj):
        """
        Get the value of the attribute configured in this attribute getter from
        the given model object.

        While resolving the attribute on the model object, a few exceptions are
        automatically caught (leading to the ``MISSING_OBJECT`` constant being
        returned):

        * ``AttributeError``, but only if an object in the attribute chain is
          None. This allows to use this getter in conjunction with nullable
          model fields.
        * ``ObjectDoesNotExist``, which is raised by Django e.g. if an object
          does not exist for reverse one-to-one relations. This allows to use
          this getter in conjunction with such relations as well.

        :param django.db.models.Model obj: The model object to retrieve the
                                           value for.
        :return: The value retrieved from the model object.
        """
        for attribute_name in self.query_path:
            obj = self._get_attribute(obj, attribute_name)
            if obj is MISSING_OBJECT:
                break
        return obj

    def get_values(self, obj):
        """
        Similar to :meth:`get_value`, but also handles m2m relations and can
        therefore return multiple values.

        The result is always a list, even if there is only one value. The
        ``MISSING_OBJECT`` constant will never be part of the resulting list
        and simply be omitted instead.

        :param django.db.models.Model obj: The model object to retrieve the
                                           values for.
        :return: A list containing the retrieved values.
        :rtype: list
        """
        values = [obj]
        for attribute_name in self.query_path:
            new_values = []
            for value in values:
                new_value = self._get_attribute(value, attribute_name)
                if isinstance(new_value, Manager):
                    new_values.extend(new_value.all())
                elif new_value is not MISSING_OBJECT:
                    new_values.append(new_value)
            values = new_values
        return values

    def build_filter(self, lookup, value):
        """
        Build a filter condition based on the configured attribute and the
        given lookup and value.

        :param str lookup: The lookup to use for the filter condition.
        :param value: The value to filter against.
        :return: The filter condition as a Q object.
        :rtype: django.db.models.Q
        """
        return (self.query_path + lookup).build_filter(value)


def parametrizable_decorator(function):
    """
    A decorator for functions who themselves are to be used as decorators and
    are to support both a parameter-less decorator usage (``@my_decorator``) as
    well as parametrized decorator usage (``@my_decorator(some_param=5)``).
    This decorator takes care of making the distinction between both usages and
    returning the correct object.

    :param function function: The decorator function to decorate.
    :return: A wrapper function that will replace the decorated function.
    :rtype: function
    """
    @wraps(function)
    def wrapper(decorated_function=None, *args, **kwargs):
        def decorator(func):
            return function(func, *args, **kwargs)

        if decorated_function:  # A function was given directly -> apply the decorator directly (@my_decorator usage).
            return decorator(decorated_function)
        return decorator  # No function -> return the actual decorator (@my_decorator(some_param=5) usage).
    return wrapper


parametrizable_decorator_method = method_decorator(parametrizable_decorator)


def get_queryable_property_descriptor(model, name):
    """
    Retrieve the descriptor object for the property with the given attribute
    name from the given model class or raise an error if no queryable property
    with that name exists on the model class.

    :param type model: The model class to retrieve the descriptor object from.
    :param str name: The name of the property to retrieve the descriptor for.
    :return: The descriptor object.
    :rtype: queryable_properties.properties.base.QueryablePropertyDescriptor
    """
    from ..properties.base import QueryablePropertyDescriptor

    descriptor = getattr(model, name, None)
    if not isinstance(descriptor, QueryablePropertyDescriptor):
        raise QueryablePropertyDoesNotExist("{model} has no queryable property named '{name}'".format(
            model=model.__name__, name=name))
    return descriptor


def resolve_queryable_property(model, query_path):
    """
    Resolve the given path into a queryable property on the given model.

    :param type model: The model to start resolving from.
    :param QueryPath query_path: The query path to resolve.
    :return: A 2-tuple containing a queryable property reference for the
             resolved property and a query path containing the parts of the
             path that represent lookups (or transforms). The first item will
             be None and the query path will be empty if no queryable property
             could be resolved.
    :rtype: (QueryablePropertyReference | None, QueryPath)
    """
    from . import get_queryable_property

    property_ref, lookups = None, QueryPath()
    # Try to follow the given path to allow to use queryable properties
    # across relations.
    for index, name in enumerate(query_path):
        try:
            related_model = get_related_model(model, name)
        except FieldDoesNotExist:
            try:
                prop = get_queryable_property(model, name)
            except QueryablePropertyDoesNotExist:
                # Neither a field nor a queryable property, so likely an
                # invalid name. Do nothing and let Django deal with it.
                pass
            else:
                property_ref = QueryablePropertyReference(prop, model, query_path[:index])
                lookups = query_path[index + 1:]
            # The current name was not a field and either a queryable
            # property or invalid. Either way, resolving ends here.
            break
        else:
            if not related_model:
                # A regular model field that doesn't represent a relation,
                # meaning that no queryable property is involved.
                break
            model = related_model
    return property_ref, lookups


def get_output_field(annotation):
    """
    Return the output field of an annotation if it can be determined.

    :param annotation: The annotation to get the output field from.
    :return: The output field of the annotation or None if it can't be
             determined.
    :rtype: django.db.models.Field | None
    """
    return getattr(annotation, 'output_field', None)



# -*- coding: utf-8 -*-

import pytest

from ..app_management.models import (
    ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties, VersionWithClassBasedProperties,
    VersionWithDecoratorBasedProperties,
)

pytestmark = [pytest.mark.asyncio, pytest.mark.django_db(transaction=True), pytest.mark.usefixtures('versions')]


@pytest.mark.parametrize('model', [ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties])
async def test_aiterator(model):
    queryset = model.objects.filter(version_count=4).select_properties('version_count')
    async for application in queryset.aiterator():
        assert model.version_count.has_cached_value(application)
        assert application.version_count == 4
    assert queryset._result_cache is None


@pytest.mark.parametrize('model', [VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties])
async def test_aupdate(model):
    queryset = model.objects.filter(major_minor='2.0')
    pks = [pk async for pk in queryset.values_list('pk', flat=True)]
    assert await queryset.aupdate(major_minor='42.42') == len(pks)
    for pk in pks:
        version = await model.objects.aget(pk=pk)  # Reload from DB
        assert version.major_minor == '42.42'



# -*- coding: utf-8 -*-

import pytest
from django import VERSION as DJANGO_VERSION

from queryable_properties.query import (
    QUERYING_PROPERTIES_MARKER, AggregatePropertyChecker, QueryablePropertiesCompilerMixin,
)
from .app_management.models import (
    ApplicationWithClassBasedProperties, CategoryWithClassBasedProperties, VersionWithClassBasedProperties,
)


class TestAggregatePropertyChecker(object):

    def test_initializer(self):
        checker = AggregatePropertyChecker()
        assert checker.func == checker.is_aggregate_property

    @pytest.mark.parametrize('model, path, value, expected_result', [
        # Not a queryable property
        (ApplicationWithClassBasedProperties, 'non_existent', None, False),
        # Queryable property with required aggregate annotation
        (ApplicationWithClassBasedProperties, 'version_count', 1337, True),
        # Queryable property without required annotation
        (VersionWithClassBasedProperties, 'version', '1.2.3', False),
        # Self-references don't lead to infinite loops
        (CategoryWithClassBasedProperties, 'circular', None, DJANGO_VERSION < (1, 8)),
    ])
    def test_is_aggregate_property(self, model, path, value, expected_result):
        assert AggregatePropertyChecker().is_aggregate_property((path, value), model) is expected_result


class TestQueryablePropertiesCompilerMixin(object):

    @pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="The setup_query method didn't exist before Django 1.8")
    def test_setup_query(self):
        queryset = ApplicationWithClassBasedProperties.objects.select_properties('version_count')
        compiler = QueryablePropertiesCompilerMixin.inject_into_object(queryset.query.get_compiler(using=queryset.db))
        compiler.setup_query()
        assert tuple(compiler.annotation_col_map) == (QUERYING_PROPERTIES_MARKER, 'version_count')



# -*- coding: utf-8 -*-
import pytest
from django import VERSION as DJANGO_VERSION
from django.db import models

from queryable_properties.properties import QueryableProperty, SubqueryExistenceCheckProperty, SubqueryFieldProperty
from queryable_properties.utils import get_queryable_property
from ..app_management.models import (
    ApplicationWithClassBasedProperties, CategoryWithClassBasedProperties, VersionWithClassBasedProperties,
)

pytestmark = [
    pytest.mark.skipif(DJANGO_VERSION < (1, 11), reason="Explicit subqueries didn't exist before Django 1.11")
]


class TestSubqueryFieldProperty(object):

    @pytest.mark.parametrize('kwargs', [
        {
            'queryset': ApplicationWithClassBasedProperties.objects.filter(name='test'),
            'field_name': 'name',
        },
        {
            'queryset': ApplicationWithClassBasedProperties.objects.all(),
            'field_name': 'common_data',
            'output_field': models.IntegerField(),
            'cached': True,
        }
    ])
    def test_initializer(self, kwargs):
        prop = SubqueryFieldProperty(**kwargs)
        assert prop.queryset is kwargs['queryset']
        assert prop.field_name == kwargs['field_name']
        assert prop.output_field is kwargs.get('output_field')
        assert prop.cached is kwargs.get('cached', QueryableProperty.cached)

    @pytest.mark.django_db
    @pytest.mark.usefixtures('versions')
    @pytest.mark.parametrize('field_name, expected_value', [
        ('version', '2.0.0'),
        ('major', 2),
    ])
    def test_build_subquery(self, monkeypatch, applications, field_name, expected_value):
        prop = get_queryable_property(ApplicationWithClassBasedProperties, 'highest_version')
        monkeypatch.setattr(prop, 'field_name', field_name)
        assert applications[0].highest_version == expected_value


class TestSubqueryExistenceCheckProperty(object):

    @pytest.mark.parametrize('kwargs', [
        {
            'queryset': ApplicationWithClassBasedProperties.objects.filter(name='test'),
        },
        {
            'queryset': ApplicationWithClassBasedProperties.objects.all(),
            'negated': True,
            'cached': True,
        }
    ])
    def test_initializer(self, kwargs):
        prop = SubqueryExistenceCheckProperty(**kwargs)
        assert prop.queryset is kwargs['queryset']
        assert prop.negated == kwargs.get('negated', False)
        assert prop.cached is kwargs.get('cached', QueryableProperty.cached)

    @pytest.mark.django_db
    @pytest.mark.usefixtures('versions')
    @pytest.mark.parametrize('negated, delete_v2, expected_result', [
        (False, False, True),
        (False, True, False),
        (True, False, False),
        (True, True, True),
    ])
    def test_build_subquery(self, monkeypatch, categories, negated, delete_v2, expected_result):
        if delete_v2:
            VersionWithClassBasedProperties.objects.filter(major=2).delete()
        prop = get_queryable_property(CategoryWithClassBasedProperties, 'has_v2')
        monkeypatch.setattr(prop, 'negated', negated)
        assert categories[0].has_v2 is expected_result



# -*- coding: utf-8 -*-

from itertools import groupby

import pytest

from queryable_properties.exceptions import QueryablePropertyDoesNotExist
from queryable_properties.properties import QueryableProperty
from queryable_properties.utils import get_queryable_property, prefetch_queryable_properties
from queryable_properties.utils.internal import get_queryable_property_descriptor
from ..app_management.models import (
    ApplicationWithClassBasedProperties, CategoryWithClassBasedProperties, VersionWithClassBasedProperties,
    VersionWithDecoratorBasedProperties,
)


class TestGetQueryableProperty(object):

    @pytest.mark.parametrize('model, property_name', [
        (VersionWithClassBasedProperties, 'major_minor'),
        (VersionWithDecoratorBasedProperties, 'major_minor'),
        (VersionWithClassBasedProperties, 'version'),
        (VersionWithDecoratorBasedProperties, 'version'),
    ])
    def test_property_found(self, model, property_name):
        prop = get_queryable_property(model, property_name)
        assert isinstance(prop, QueryableProperty)

    @pytest.mark.parametrize('model, property_name', [
        (VersionWithClassBasedProperties, 'non_existent'),
        (VersionWithDecoratorBasedProperties, 'non_existent'),
        (VersionWithClassBasedProperties, 'major'),  # Existing model field
        (VersionWithDecoratorBasedProperties, 'major'),  # Existing model field
    ])
    def test_exception(self, model, property_name):
        with pytest.raises(QueryablePropertyDoesNotExist):
            get_queryable_property(model, property_name)


@pytest.mark.django_db
class TestPrefetchQueryableProperties(object):

    def assert_not_cached(self, descriptor, *model_instances):
        assert all(not descriptor.has_cached_value(instance) for instance in model_instances)

    def assert_cached(self, descriptor, *model_instances):
        assert all(descriptor.has_cached_value(instance) for instance in model_instances)
        assert all(descriptor.get_cached_value(instance) == descriptor.prop.get_value(instance)
                   for instance in model_instances)

    def test_no_instances(self):
        instances = []
        prefetch_queryable_properties(instances, 'whatever')
        assert instances == []

    @pytest.mark.parametrize('property_names', [
        (),
        ('version_count',),
        ('version_count', 'major_sum'),
    ])
    @pytest.mark.usefixtures('versions')
    def test_local_property(self, applications, property_names):
        descriptors = [get_queryable_property_descriptor(ApplicationWithClassBasedProperties, name)
                       for name in property_names]
        applications = applications[:2]
        for descriptor in descriptors:
            self.assert_not_cached(descriptor, *applications)
        prefetch_queryable_properties(applications, *property_names)
        for descriptor in descriptors:
            self.assert_cached(descriptor, *applications)

    @pytest.mark.usefixtures('versions')
    def test_2o_relation_property(self):
        descriptor = get_queryable_property_descriptor(ApplicationWithClassBasedProperties, 'version_count')
        versions = list(VersionWithClassBasedProperties.objects.select_related('application'))
        self.assert_not_cached(descriptor, *(version.application for version in versions))
        prefetch_queryable_properties(versions, 'application__version_count')
        self.assert_cached(descriptor, *(version.application for version in versions))

    @pytest.mark.usefixtures('versions')
    def test_2m_relation_property(self):
        descriptor = get_queryable_property_descriptor(ApplicationWithClassBasedProperties, 'version_count')
        categories = list(CategoryWithClassBasedProperties.objects.prefetch_related('applications'))
        for category in categories:
            self.assert_not_cached(descriptor, *category.applications.all())
        prefetch_queryable_properties(categories, 'applications__version_count')
        for category in categories:
            self.assert_cached(descriptor, *category.applications.all())

    @pytest.mark.usefixtures('versions')
    def test_different_types(self, categories, applications):
        model_instances = categories[:2] + applications
        grouped_instances = {cls: list(instances) for cls, instances in
                             groupby(model_instances, key=lambda instance: instance.__class__)}
        for cls, instances in grouped_instances.items():
            self.assert_not_cached(get_queryable_property_descriptor(cls, 'version_count'), *instances)
        prefetch_queryable_properties(model_instances, 'version_count')
        for cls, instances in grouped_instances.items():
            self.assert_cached(get_queryable_property_descriptor(cls, 'version_count'), *instances)

    @pytest.mark.usefixtures('versions')
    def test_refresh_cache(self):
        """
        Test that prefetch_queryable_properties can be used to update already
        cached values on instances.
        """
        descriptor = get_queryable_property_descriptor(ApplicationWithClassBasedProperties, 'version_count')
        application = ApplicationWithClassBasedProperties.objects.select_properties('version_count')[0]
        self.assert_cached(descriptor, application)
        application.versions.all().delete()
        assert application.version_count > 0  # The cached value is still present
        prefetch_queryable_properties([application], 'version_count')
        self.assert_cached(descriptor, application)

    @pytest.mark.parametrize('property_name, expected_exception', [
        ('non_existent', QueryablePropertyDoesNotExist),
        ('application__non_existent', QueryablePropertyDoesNotExist),
        ('non_existent__non_existent', AttributeError),
    ])
    def test_invalid_propery_path(self, versions, property_name, expected_exception):
        with pytest.raises(expected_exception):
            prefetch_queryable_properties(versions, property_name)



# -*- coding: utf-8 -*-

import pytest
import six
from django import VERSION as DJANGO_VERSION
from django.contrib.admin import ModelAdmin, SimpleListFilter, site
from django.contrib.admin.filters import AllValuesFieldListFilter
from django.core.exceptions import ImproperlyConfigured
from django.db.models import F

from queryable_properties.admin.checks import Error
from queryable_properties.compat import admin_validation
from ..app_management.admin import ApplicationAdmin, VersionAdmin, VersionInline
from ..app_management.models import ApplicationWithClassBasedProperties, VersionWithClassBasedProperties
from ..conftest import Concat, Value


class Dummy(object):

    def __str__(self):
        return self.__class__.__name__


class DummyListFilter(SimpleListFilter):
    parameter_name = 'dummy'


def assert_admin_validation(admin_class, model, error_id=None, exception_text=None):
    """
    Validate an admin class and compare the result to the given expectation.

    :param admin_class: The admin class to validate (may be an inline class).
    :param model: The model class the admin class is intended for.
    :param str | None error_id: The expected error ID, which is used for new
                                Django versions. A value of None means that
                                the validation is expected to not find any
                                errors.
    :param str | None exception_text: The expected error text, which is used
                                for old Django versions. A value of None means
                                that the validation is expected to not find any
                                errors.
    """
    if hasattr(ModelAdmin, 'check'):
        if DJANGO_VERSION >= (1, 9):
            errors = admin_class(model, site).check()
        else:
            errors = admin_class.check(model)
        if error_id is None:
            assert not errors
        else:
            assert any(error.id == error_id for error in errors)

    if hasattr(admin_validation, 'validate') or hasattr(ModelAdmin, 'validate'):
        try:
            if hasattr(ModelAdmin, 'validate'):
                admin_class.validate(model)
            else:
                admin_validation.validate(admin_class, model)
        except ImproperlyConfigured as e:
            assert exception_text is not None
            assert exception_text in six.text_type(e)
        else:
            assert exception_text is None


class TestError(object):

    def test_initializer(self):
        error = Error('test message', Dummy, 42)
        assert error.msg == 'test message'
        assert error.obj is Dummy
        assert error.id == 'queryable_properties.admin.E042'

    def test_raise_exception(self):
        error = Error('test message', Dummy(), 42)
        with pytest.raises(ImproperlyConfigured, match=r'Dummy: \(queryable_properties\.admin\.E042\) test message'):
            error.raise_exception()


class TestQueryablePropertiesChecksMixin(object):

    @pytest.mark.parametrize('admin, model', [
        (VersionAdmin, VersionWithClassBasedProperties),
        (ApplicationAdmin, ApplicationWithClassBasedProperties),
    ])
    def test_success(self, admin, model):
        assert_admin_validation(admin, model)

    @pytest.mark.parametrize('filter_item', [
        DummyListFilter,
        ('common_data', AllValuesFieldListFilter),
        ('support_start_date', AllValuesFieldListFilter),
    ])
    def test_list_filter_valid_items(self, monkeypatch, filter_item):
        monkeypatch.setattr(ApplicationAdmin, 'list_filter', (filter_item,))
        assert_admin_validation(ApplicationAdmin, ApplicationWithClassBasedProperties)

    @pytest.mark.parametrize('filter_item, error_id', [
        ('major_minor', 'queryable_properties.admin.E002'),
        ('is_supported__isnull', 'queryable_properties.admin.E004'),
    ])
    def test_list_filter_invalid_property(self, monkeypatch, filter_item, error_id):
        monkeypatch.setattr(VersionAdmin, 'list_filter', (filter_item,))
        assert_admin_validation(VersionAdmin, VersionWithClassBasedProperties, error_id, '({})'.format(error_id))

    @pytest.mark.parametrize('filter_item, error_id, exception_text', [
        (Dummy, 'admin.E113', "'Dummy' which is not a descendant of ListFilter"),
        (AllValuesFieldListFilter, 'admin.E114', "'AllValuesFieldListFilter' which is of type FieldListFilter"),
        (('version_count', DummyListFilter), 'admin.E115', "'DummyListFilter' which is not of type FieldListFilter"),
        ('neither_property_nor_field', 'admin.E116', "'neither_property_nor_field' which does not refer to a Field"),
    ])
    def test_list_filter_invalid_items(self, monkeypatch, filter_item, error_id, exception_text):
        monkeypatch.setattr(ApplicationAdmin, 'list_filter', (filter_item,))
        assert_admin_validation(ApplicationAdmin, ApplicationWithClassBasedProperties, error_id, exception_text)

    @pytest.mark.parametrize('admin_class', [ApplicationAdmin, VersionInline])
    def test_list_select_properties_invalid_type(self, monkeypatch, admin_class):
        monkeypatch.setattr(admin_class, 'list_select_properties', None)
        assert_admin_validation(ApplicationAdmin, ApplicationWithClassBasedProperties,
                                'queryable_properties.admin.E005', '(queryable_properties.admin.E005)')

    @pytest.mark.parametrize('admin_class, property_name, error_id', [
        (ApplicationAdmin, 'name', 'queryable_properties.admin.E001'),
        (VersionInline, 'major', 'queryable_properties.admin.E001'),
        (ApplicationAdmin, 'dummy', 'queryable_properties.admin.E002'),
        (VersionInline, 'major_minor', 'queryable_properties.admin.E002'),
        (ApplicationAdmin, 'categories__version_count', 'queryable_properties.admin.E003'),
        (VersionInline, 'application__categories__version_count', 'queryable_properties.admin.E003'),
        (ApplicationAdmin, 'version_count__lt', 'queryable_properties.admin.E004'),
        (VersionInline, 'version__regex', 'queryable_properties.admin.E004'),
    ])
    def test_list_select_properties_invalid_property(self, monkeypatch, admin_class, property_name, error_id):
        monkeypatch.setattr(admin_class, 'list_select_properties', (property_name,))
        assert_admin_validation(ApplicationAdmin, ApplicationWithClassBasedProperties,
                                error_id, '({})'.format(error_id))

    @pytest.mark.parametrize('admin_class, property_name', [
        (ApplicationAdmin, 'version_count'),
        (VersionInline, 'version'),
    ])
    def test_ordering_valid_desc(self, monkeypatch, admin_class, property_name):
        monkeypatch.setattr(admin_class, 'ordering', ('-' + property_name,))
        assert_admin_validation(ApplicationAdmin, ApplicationWithClassBasedProperties)

    @pytest.mark.skipif(DJANGO_VERSION < (2, 0), reason="Expression-based ordering wasn't supported before Django 2.0")
    @pytest.mark.parametrize('admin_class, expression', [
        (ApplicationAdmin, F('highest_version')),
        (VersionInline, F('version')),
        (ApplicationAdmin, Concat(Value('V'), 'highest_version')),
        (VersionInline, Concat(Value('V'), 'version')),
        (ApplicationAdmin, Concat(Value('V'), 'highest_version').desc()),
        (VersionInline, Concat(Value('V'), 'version').desc()),
    ])
    def test_ordering_valid_expressions(self, monkeypatch, admin_class, expression):
        monkeypatch.setattr(admin_class, 'ordering', (expression,))
        assert_admin_validation(ApplicationAdmin, ApplicationWithClassBasedProperties)

    @pytest.mark.parametrize('admin_class, property_name', [
        (ApplicationAdmin, 'dummy'),
        (VersionInline, 'major_minor'),
    ])
    def test_ordering_invalid_property(self, monkeypatch, admin_class, property_name):
        monkeypatch.setattr(admin_class, 'ordering', (property_name,))
        assert_admin_validation(ApplicationAdmin, ApplicationWithClassBasedProperties,
                                'queryable_properties.admin.E002', '(queryable_properties.admin.E002)')



# -*- coding: utf-8 -*-

from django.contrib.admin import ModelAdmin, StackedInline, TabularInline

from ..compat import ADMIN_QUERYSET_METHOD_NAME, admin_validation
from ..exceptions import QueryablePropertyError
from ..managers import QueryablePropertiesQuerySetMixin
from ..utils.internal import QueryPath
from .checks import QueryablePropertiesChecksMixin
from .filters import QueryablePropertyField

__all__ = [
    'QueryablePropertiesAdmin',
    'QueryablePropertiesAdminMixin',
    'QueryablePropertiesStackedInline',
    'QueryablePropertiesTabularInline',
]


class QueryablePropertiesAdminMixin(object):
    """
    A mixin for admin classes including inlines that allows to use queryable
    properties in various admin features.
    """

    list_select_properties = ()
    """A sequence of queryable property names that should be selected."""

    def __init__(self, *args, **kwargs):
        super(QueryablePropertiesAdminMixin, self).__init__(*args, **kwargs)
        if hasattr(self, 'list_filter') and not hasattr(ModelAdmin, 'get_list_filter'):  # pragma: no cover
            # In very old Django versions, there was no get_list_filter method,
            # therefore the processed queryable property filters must be stored
            # directly in the list_filter attribute.
            self.list_filter = self.process_queryable_property_filters(self.list_filter)

    @classmethod
    def validate(cls, model):  # pragma: no cover
        cls._ensure_queryable_property_checks()
        return super(QueryablePropertiesAdminMixin, cls).validate(model)

    def check(self, *args, **kwargs):
        self._ensure_queryable_property_checks(self)
        return super(QueryablePropertiesAdminMixin, self).check(*args, **kwargs)

    if getattr(getattr(ModelAdmin, 'check', None), '__self__', None):  # pragma: no cover
        # In old Django versions, check was a classmethod.
        check = classmethod(check)

    @classmethod
    def _ensure_queryable_property_checks(cls, obj=None):
        """
        Make sure that the queryable properties admin check extensions are used
        to avoid errors due to Django's default validation, which would treat
        queryable property names as invalid.

        :param obj: The (optional) model admin instance to ensure the queryable
                    property checks for. If not provided, they are ensured for
                    the current class instead.
        :type obj: ModelAdmin | None
        """
        obj = obj or cls
        # Dynamically add a mixin that handles queryable properties into the
        # admin's checks/validation class.
        for attr_name in ('checks_class', 'validator_class', 'default_validator_class'):
            checks_class = getattr(obj, attr_name, None)
            if checks_class:
                class_name = 'QueryableProperties' + checks_class.__name__
                setattr(obj, attr_name, QueryablePropertiesChecksMixin.mix_with_class(checks_class, class_name))

    def get_queryset(self, request):
        # The base method has different names in different Django versions (see
        # comment on the constant definition).
        base_method = getattr(super(QueryablePropertiesAdminMixin, self), ADMIN_QUERYSET_METHOD_NAME)
        # Make sure to use a queryset with queryable properties features.
        queryset = QueryablePropertiesQuerySetMixin.apply_to(base_method(request))
        # Apply list_select_properties.
        list_select_properties = self.get_list_select_properties(request)
        if list_select_properties:
            queryset = queryset.select_properties(*list_select_properties)
        return queryset

    def queryset(self, request):  # pragma: no cover
        # Same as get_queryset, but for very old Django versions. Simply
        # delegate to get_queryset, which is aware of the different methods in
        # different versions and therefore calls the correct super methods if
        # necessary.
        return self.get_queryset(request)

    def get_list_select_properties(self, request):
        """
        Wrapper around the ``list_select_properties`` attribute that allows to
        dynamically create the list of queryable property names to select based
        on the given request.

        :param django.http.HttpRequest request: The request to the admin.
        :return: A sequence of queryable property names to select.
        :rtype: collections.Sequence[str]
        """
        return self.list_select_properties

    def get_list_filter(self, request):
        list_filter = super(QueryablePropertiesAdminMixin, self).get_list_filter(request)
        return self.process_queryable_property_filters(list_filter)

    def process_queryable_property_filters(self, list_filter):
        """
        Process a sequence of list filters to create a new sequence in which
        queryable property references are replaced with custom callables that
        make them compatible with Django's filter workflow.

        :param collections.Sequence list_filter: The list filter sequence.
        :return: The processed list filter sequence.
        :rtype: list
        """
        processed_filters = []
        for item in list_filter:
            if not callable(item):
                if isinstance(item, (tuple, list)):
                    field_name, filter_class = item
                else:
                    field_name, filter_class = item, None
                try:
                    item = QueryablePropertyField(self, QueryPath(field_name)).get_filter_creator(filter_class)
                except QueryablePropertyError:
                    pass
            processed_filters.append(item)
        return processed_filters


class QueryablePropertiesAdmin(QueryablePropertiesAdminMixin, ModelAdmin):
    """
    Base class for admin classes which allows to use queryable properties in
    various admin features.

    Intended to be used in place of Django's regular ``ModelAdmin`` class.
    """


class QueryablePropertiesStackedInline(QueryablePropertiesAdminMixin, StackedInline):
    """
    Base class for stacked inline classes which allows to use queryable
    properties in various admin features.

    Intended to be used in place of Django's regular ``StackedInline`` class.
    """


class QueryablePropertiesTabularInline(QueryablePropertiesAdminMixin, TabularInline):
    """
    Base class for tabular inline classes which allows to use queryable
    properties in various admin features.

    Intended to be used in place of Django's regular ``TabularInline`` class.
    """


# In very old django versions, the admin validation happens in one big function
# that cannot really be extended well. Therefore, the Django module will be
# monkeypatched in order to allow the queryable properties validation to take
# effect.
django_validate = getattr(admin_validation, 'validate', None)
django_validate_inline = getattr(admin_validation, 'validate_inline', None)

if django_validate:  # pragma: no cover
    def validate(cls, model):
        if issubclass(cls, QueryablePropertiesAdminMixin):
            cls = QueryablePropertiesChecksMixin()._validate_queryable_properties(cls, model)
        django_validate(cls, model)

    admin_validation.validate = validate

if django_validate_inline:  # pragma: no cover
    def validate_inline(cls, parent, parent_model):
        if issubclass(cls, QueryablePropertiesAdminMixin):
            cls = QueryablePropertiesChecksMixin()._validate_queryable_properties(cls, cls.model)
        django_validate_inline(cls, parent, parent_model)

    admin_validation.validate_inline = validate_inline
